# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Madara – WordPress Theme for Manga

# Supported servers:
# 24hRomance [EN] (disabled)
# 3asq [AR] (disabled)
# AkuManga [AR] (disabled)
# Aloalivn [EN] (disabled)
# Apoll Comics [ES]
# ArazNovel [TR] (disabled)
# Argos Scan / Argos Comics [PT] (disabled)
# Asura Scans [TR]
# Atikrost [TR] (disabled)
# Best Manga [RU] (disabled)
# Fr-Scan (Id frdashscan) [FR] (disabled)
# Hunlight Scans [EN] (disabled)
# Leomanga [ES]
# Leviatanscans [EN] (disabled)
# Lovers Toon [pt_BR]
# Manga-Scantrad [FR]
# Mangas Origines [FR]
# MangaWeebs [EN] (disabled)
# Manhuaus [EN]
# Manhwa Hentai [EN] (disabled)
# Phoenix Fansub [ES] (disabled)
# Poseidon Scans [FR] (disabled)
# Reaperscans [AR/FR/ID/TR] (disabled)
# Starbound Scans [FR] (disabled)
# Submanga [ES] (disabled)
# ToonGod [EN]
# Toonily [EN]
# Wakascan [FR] (disabled)
# Webtoon Hatti [TR]

import datetime
from gettext import gettext as _
import logging
import re

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.models import Settings
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import get_soup_element_inner_text
from komikku.servers.utils import remove_emoji_from_string
from komikku.utils import get_buffer_mime_type
from komikku.webview import CompleteChallenge

logger = logging.getLogger('komikku.servers.madara')


class Madara(Server):
    base_url: str = None
    chapter_url: str = None
    chapters_url: str = None

    date_format: str = '%B %d, %Y'
    medium: str = 'manga'
    series_name: str = 'manga'

    results_selector = '.row'
    result_name_slug_selector = '.post-title a'
    result_cover_selector = '.tab-thumb img'

    details_name_selector = 'h1'
    details_cover_selector = '.summary_image > a > img'
    details_authors_selector = '.author-content a, .artist-content a'
    details_scanlators_selector = None
    details_genres_selector = '.genres-content a'
    details_status_selector = '.post-status .summary-content'
    details_synopsis_selector = '.summary__content'

    chapters_list_selector = '#manga-chapters-holder'
    chapters_date_selector = '.chapter-release-date'
    chapters_selector = '.wp-manga-chapter'
    chapters_slug_range = (-2, -1)
    chapters_order = 'desc'

    images_src_attr = 'src'

    def __init__(self):
        self.api_url = self.base_url + '/wp-admin/admin-ajax.php'
        self.manga_url = self.base_url + '/' + self.series_name + '/{0}/'
        if self.chapter_url is None:
            self.chapter_url = self.base_url + '/' + self.series_name + '/{0}/{1}/?style=list'

        if self.session is None and not self.has_cf:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    @staticmethod
    def extract_chapter_nums_from_slug(slug):
        re_nums = r'((\w+)\-(\d+)-)?(\w+)-(\d+)([-_](\d+))?.*'

        if matches := re.search(re_nums, slug):
            if num := matches.group(5):
                num = f'{int(num)}'

                if num_dec := matches.group(7):
                    num = f'{num}.{int(num_dec)}'

                if num_volume := matches.group(3):
                    num_volume = f'{int(num_volume)}'

                return num, num_volume

        return None, None

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(
            self.manga_url.format(initial_data['slug']),
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type not in ('text/html', 'text/plain'):
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))
        if r.history and r.history[-1].status_code == 301:
            # Slug has changed
            data['slug'] = r.url.split('/')[-2]

        data['name'] = get_soup_element_inner_text(soup.select_one(self.details_name_selector))
        if cover_element := soup.select_one(self.details_cover_selector):
            data['cover'] = cover_element.get('data-src')
            if data['cover'] is None:
                data['cover'] = cover_element.get('data-lazy-src')
                if data['cover'] is None:
                    data['cover'] = cover_element.get('data-lazy-srcset')
                    if data['cover']:
                        # data-lazy-srcset can contain several covers with sizes: url1 size1 url2 size2...
                        data['cover'] = data['cover'].split()[0]
                    else:
                        data['cover'] = cover_element.get('src')

        # Details
        for element in soup.select(self.details_authors_selector):
            author = element.text.strip()
            if author not in data['authors']:
                data['authors'].append(author)

        if self.details_scanlators_selector:
            for element in soup.select(self.details_scanlators_selector):
                data['scanlators'].append(element.text.strip())

        for element in soup.select(self.details_genres_selector):
            genre = element.text.strip()
            if genre not in data['genres']:
                data['genres'].append(genre)

        if self.details_status_selector:
            if elements := soup.select(self.details_status_selector):
                element = elements[-1]  # get last element (release date is sometimes shown just before)
                status = element.text.strip()
                # Remove emoji
                status = remove_emoji_from_string(status)

                if status in ('Completed', 'Terminé', 'Completé', 'Completo', 'Finalizado', 'Terminado', 'Concluído', 'Tamamlandı', 'مكتملة', 'Закончена'):
                    data['status'] = 'complete'
                elif status in ('OnGoing', 'En Cours', 'En cours', 'Updating', 'Devam Ediyor', 'Em Lançamento', 'Em andamento', 'En curso', 'En Emision', 'En emisión', 'En Emisión', 'مستمرة', 'Продолжается', 'Выпускается'):
                    data['status'] = 'ongoing'
                elif status in ('Canceled', 'Cancelada', 'Cancelado'):
                    data['status'] = 'suspended'
                elif status in ('On Hold', 'En pause', 'En Hiatus', 'En espera'):
                    data['status'] = 'hiatus'

        if self.details_synopsis_selector:
            if summary_container := soup.select_one(self.details_synopsis_selector):
                if p_elements := summary_container.select('p'):
                    synopsis = []
                    for p_element in p_elements:
                        if paragraph := p_element.text.strip():
                            synopsis.append(paragraph)
                    data['synopsis'] = '\n\n'.join(synopsis) if synopsis else None
                else:
                    data['synopsis'] = summary_container.text.strip()

        # Chapters
        chapters_container = soup.select_one(self.chapters_list_selector)
        if chapters_container:
            # Chapters list is empty and is loaded via an Ajax call
            if self.chapters_url:
                r = self.session_post(
                    self.chapters_url.format(data['slug']),
                    headers={
                        'Origin': self.base_url,
                        'Referer': self.manga_url.format(data['slug']),
                        'X-Requested-With': 'XMLHttpRequest',
                    }
                )
            else:
                r = self.session_post(
                    self.api_url,
                    data=dict(
                        action='manga_get_chapters',
                        manga=chapters_container.get('data-id'),
                    ),
                    headers={
                        'Origin': self.base_url,
                        'Referer': self.manga_url.format(data['slug']),
                        'X-rRquested-With': 'XMLHttpRequest',
                    }
                )

            soup = BeautifulSoup(r.text, 'lxml')

        elements = soup.select(self.chapters_selector)
        if self.chapters_order == 'desc':
            elements = reversed(elements)

        for element in elements:
            if element.select_one('i.fa-lock'):
                # Skip premium chapter (LeviatanScans ES for ex.)
                continue

            a_element = element.a
            if view_element := element.find(class_='view'):
                view_element.extract()

            url = a_element.get('href')
            if 'javascript' in url:
                # VIP
                continue

            date_element = element.select_one(self.chapters_date_selector).extract()
            if date := date_element.text.strip():
                date = convert_date_string(date, format=self.date_format, languages=[self.lang])
            else:
                date = datetime.date.today().strftime('%Y-%m-%d')

            slug = url.split('/')[-2]
            num, num_volume = self.extract_chapter_nums_from_slug(slug)

            data['chapters'].append(dict(
                slug=slug,
                title=a_element.text.strip(),
                num=num,
                num_volume=num_volume,
                date=date,
            ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for index, img_element in enumerate(soup.select_one('.read-container, .reading-content, .image-list').select('img.wp-manga-chapter-img')):
            if img_element.parent.name == 'noscript':
                # In case server uses a second <img> encapsulated in a <noscript> element
                continue

            img_url = img_element.get(self.images_src_attr)
            if img_url is None:
                continue

            data['pages'].append(dict(
                slug=None,
                image=img_url.strip(),
                index=index + 1,
            ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0:03d}.{1}'.format(page['index'], mime_type.split('/')[-1]),
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        """
        Returns list of latest updates manga
        """
        return self.search('', orderby='latest')

    def get_most_populars(self):
        """
        Returns list of most viewed manga
        """
        return self.search('', orderby='populars')

    @CompleteChallenge()
    def search(self, term, orderby=None):
        data = {
            'action': 'madara_load_more',
            'page': 0,
            'template': 'madara-core/content/content-search',
            'vars[orderby]': 'meta_value_num' if orderby else '',
            'vars[paged]': 0,
            'vars[template]': 'search',
            'vars[post_type]': 'wp-manga',
            'vars[post_status]': 'publish',
            'vars[manga_archives_item_layout]': 'default',
        }

        if self.medium:
            data['vars[meta_query][0][0][value]'] = self.medium  # allows to ignore novels
            data['vars[meta_query][0][orderby]'] = ''
            data['vars[meta_query][0][paged]'] = '0'
            data['vars[meta_query][0][template]'] = 'search'
            data['vars[meta_query][0][meta_query][relation]'] = 'AND'
            data['vars[meta_query][0][post_type]'] = 'wp-manga'
            data['vars[meta_query][0][post_status]'] = 'publish'
            data['vars[meta_query][relation]'] = 'AND'

        if orderby:
            data['vars[order]'] = 'desc'
            data['vars[posts_per_page]'] = 100
            if orderby == 'populars':
                data['vars[meta_key]'] = '_wp_manga_views'
            elif orderby == 'latest':
                data['vars[meta_key]'] = '_latest_update'
        else:
            data['vars[meta_query][0][s]'] = term
            data['vars[s]'] = term

        r = self.session_post(
            self.api_url,
            data=data,
            headers={
                'X-Requested-With': 'XMLHttpRequest',
                'Referer': self.base_url
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select(self.results_selector):
            a_element = element.select_one(self.result_name_slug_selector)
            slug = a_element.get('href').split('/')[-2]
            name = a_element.text.strip()
            if not name or not slug:
                continue

            if cover_img := element.select_one(self.result_cover_selector):
                cover = cover_img.get('data-src')
                if cover is None:
                    cover = cover_img.get('src')

            results.append(dict(
                slug=slug,
                name=name,
                cover=cover,
            ))

        return results


class Madara2(Madara):
    filters = [
        {
            'key': 'nsfw',
            'type': 'checkbox',
            'name': _('NSFW Content'),
            'description': _('Whether to show manga containing NSFW content'),
            'default': False,
        },
    ]

    def __init__(self):
        super().__init__()

        # Update NSFW filter default value according to current settings
        if Settings.instance:
            self.filters[0]['default'] = Settings.get_default().nsfw_content

    @CompleteChallenge()
    def get_latest_updates(self, nsfw):
        """
        Returns list of latest updates manga
        """
        return self.search(None, nsfw=nsfw, orderby='latest')

    @CompleteChallenge()
    def get_most_populars(self, nsfw):
        """
        Returns list of most viewed manga
        """
        return self.search(None, nsfw=nsfw, orderby='populars')

    @CompleteChallenge()
    def search(self, term, nsfw, orderby=None):
        params = {
            's': term or '',
            'post_type': 'wp-manga',
            'op': '',
            'author': '',
            'artist': '',
            'release': '',
            'adult': '' if nsfw else 0,
        }

        if orderby == 'populars':
            params['m_orderby'] = 'views'
        elif orderby == 'latest':
            params['m_orderby'] = 'latest'

        r = self.session_get(f'{self.base_url}/', params=params)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.c-tabs-item__content'):
            thumb_element = element.select_one('.tab-thumb')
            if not thumb_element.a:
                continue

            last_chapter_element = element.select_one('.latest-chap .chapter')
            if cover_element := element.a.img or element.img:
                cover = cover_element.get('data-src')
                if not cover:
                    cover = cover_element.get('src')
            else:
                cover = None

            results.append(dict(
                slug=thumb_element.a.get('href').split('/')[-2],
                name=thumb_element.a.get('title').strip(),
                cover=cover,
                last_chapter=last_chapter_element.text.strip() if last_chapter_element else None,
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.manga_stream import MangaStream


class Shadowmangas(MangaStream):
    id = 'shadowmangas'
    name = 'ShadowMangas'
    lang = 'es'
    is_nsfw = True

    base_url = 'https://shadowmangas.com'
    logo_url = 'https://i3.wp.com/shadowmangas.com/wp-content/uploads/2022/09/cropped-icoweb-32x32.png'

    authors_selector = '.infox .fmed:-soup-contains("Author") span, .infox .fmed:-soup-contains("Artist") span'
    genres_selector = '.mgen a'
    scanlators_selector = None
    status_selector = '.tsinfo .imptdt:-soup-contains("Status") i'
    synopsis_selector = '[itemprop="description"]'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from datetime import datetime

import json
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_mri_data_to_webp_buffer
from komikku.utils import get_buffer_mime_type

SERVER_NAME = 'Manga Rock'

headers = {
    'Origin': 'https://mangarock.com',
    'User-Agent': USER_AGENT,
}

#
# BEWARE: This server is disabled
# Dead since 01/2020
#


class Mangarock(Server):
    id = 'mangarock'
    name = SERVER_NAME
    lang = 'en'
    status = 'disabled'

    base_url = 'https://mangarock.com'
    api_url = 'https://api.mangarockhd.com/query/web401'
    api_search_url = api_url + '/mrs_search?country='
    api_most_populars_url = api_url + '/mrs_latest'
    api_manga_url = api_url + '/info?oid={0}&last=0'
    # api_chapter_url = api_url + '/pages?oid={0}'
    api_chapter_url = api_url + '/pagesv2?oid={0}'
    manga_url = base_url + '/manga/{0}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update(headers)

    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.api_manga_url.format(initial_data['slug']))
        if r is None:
            return None

        try:
            res = r.json()
        except json.decoder.JSONDecodeError:
            return None

        if res['code'] != 0:
            return None

        res = res['data']

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        # Name & cover
        data['name'] = res['name']
        data['cover'] = res['thumbnail']

        # Details & Synopsis
        for author in res['authors'] or []:
            data['authors'].append('{0} ({1})'.format(author['name'], author['role']))

        for genre in res['rich_categories'] or []:
            data['genres'].append(genre['name'])

        data['status'] = 'complete' if res['completed'] else 'ongoing'

        data['synopsis'] = res['description']

        # Chapters
        for chapter in res['chapters']:
            data['chapters'].append(dict(
                slug=chapter['oid'],
                title=chapter['name'],
                date=datetime.fromtimestamp(chapter['updatedAt']).date(),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data

        Currently, only pages are expected.
        """
        r = self.session_get(self.api_chapter_url.format(chapter_slug))
        if r is None:
            return None

        try:
            res = r.json()
        except json.decoder.JSONDecodeError:
            return None

        data = dict(
            pages=[],
        )

        if res['code'] == 0:
            for page in res['data']:
                data['pages'].append(dict(
                    slug=None,  # not necessary, we know image url already
                    image=page['url'],
                ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r is None or r.status_code != 200:
            return None

        buffer = r.content
        mime_type = get_buffer_mime_type(buffer)
        if mime_type == 'application/octet-stream':
            buffer = convert_mri_data_to_webp_buffer(buffer)
            mime_type = get_buffer_mime_type(buffer)

        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=buffer,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        """
        Returns full list of manga sorted by rank
        """
        r = self.session_post(self.api_most_populars_url)
        if r is None or r.status_code != 200:
            return None

        try:
            res = r.json()
        except json.decoder.JSONDecodeError:
            return None

        if res['code'] != 0:
            return None

        # Sort by rank
        res = sorted(res['data'], key=lambda i: i['rank'])

        results = []
        for item in res:
            results.append(dict(
                name=item['name'],
                slug=item['oid'],
            ))

        return results

    def search(self, term):
        r = self.session_post(self.api_search_url, json={'type': 'series', 'keywords': term})
        if r is None or r.status_code != 200:
            return None

        try:
            res = r.json()
        except json.decoder.JSONDecodeError:
            return None

        if res['code'] != 0:
            return None

        # Returned data for each manga:
        # oid: slug of the manga
        results = []
        for oid in res['data']:
            data = self.get_manga_data(dict(slug=oid))
            if data:
                results.append(dict(
                    name=data['name'],
                    slug=data['slug'],
                ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara


class Aloalivn(Madara):
    id = 'aloalivn'
    name = 'Aloalivn'
    lang = 'en'
    status = 'disabled'

    base_url = 'https://aloalivn.com'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.fuzzydoodle import FuzzyDoodle


class Fleksyscans(FuzzyDoodle):
    id = 'fleksyscans'
    name = 'FleksyScans'
    lang = 'en'
    is_nsfw_only = True
    status = 'disabled'  # Dead 06/2024 DMCA

    base_url = 'https://flexscans.com'
    search_url = base_url + '/manga'
    manga_url = base_url + '/manga/{0}'
    chapter_url = base_url + '/manga/{0}/{1}'

    # Selectors
    search_results_selector = '#card-real'
    search_link_selector = 'a'
    search_cover_img_selector = 'a > figure > img'

    most_popular_results_selector = '#popular-cards #card-real'  # does not currently exist 04/2024
    most_popular_link_selector = 'a'
    most_popular_cover_img_selector = 'a > figure > img'

    latest_updates_results_selector = '#latest > section > div > div'
    latest_updates_link_selector = '#card-real a'
    latest_updates_cover_img_selector = '#card-real a > figure > img'
    latest_updates_last_chapter_selector = 'div:last-child > a > div > b'

    details_name_selector = 'h1'
    details_cover_selector = 'main > section > div > div > div.relative > img'
    details_status_selector = 'main > section > div > div > div > div > p:-soup-contains("Status") > a > span'
    details_author_selector = 'main > section > div > div > div > div > p:-soup-contains("Author") > span:last-child'
    details_artist_selector = 'main > section > div > div > div > div > p:-soup-contains("Artist") > span:last-child'
    details_type_selector = 'main > section > div > div > div > div > p:-soup-contains("Type") > a > span'
    details_genres_selector = 'main > section > div > div:nth-child(2) > div > div:last-child > a'
    details_synopsis_selector = 'main > section > div > div:nth-child(2) > div:nth-child(2) > div > p'

    chapters_selector = '#chapters-list > a'
    chapters_title_selector = '#item-title'
    chapters_date_selector = 'div > span:last-child'

    chapter_pages_selector = '#chapter-container img'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara2


class Mangabin(Madara2):
    id = 'mangabin'
    name = 'MangaBin'
    lang = 'en'
    is_nsfw = True
    status = 'disabled'

    base_url = 'https://mangabin.com'
    logo_url = base_url + '/wp-content/uploads/2024/12/cropped-coollogo_com-29007530-32x32.png'
    chapters_url = base_url + '/manga/{0}/ajax/chapters/'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara2


class Leomanga(Madara2):
    id = 'leomanga'
    name = 'Lectormanga (Leomanga)'
    lang = 'es'
    is_nsfw = True
    has_cf = True

    series_name = 'biblioteca'

    base_url = 'https://lectormangaa.com'



# SPDX-FileCopyrightText: 2025-2025 Lélahel Hideux
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Lélahel Hideux <lelahx@proton.me>

from komikku.servers.multi.manga_stream import MangaStream


class Sushiscannet(MangaStream):
    id = 'sushiscannet'
    name = 'Sushiscan (.net)'
    lang = 'fr'

    has_cf = True

    base_url = 'https://sushiscan.net'

    series_name = 'catalogue'

    authors_selector = '.infotable tr:-soup-contains("Dessinateur") td:last-child, .infotable tr:-soup-contains("Auteur") td:last-child'
    genres_selector = '.seriestugenre a'
    scanlators_selector = None
    status_selector = '.infotable tr:-soup-contains("Statut") td:last-child'
    synopsis_selector = '[itemprop="description"]'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from bs4 import BeautifulSoup

from komikku.servers.multi.manga_stream import MangaStream
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.webview import CompleteChallenge


class Rawmanga(MangaStream):
    id = 'rawmanga'
    name = 'Raw Manga 生漫画'
    lang = 'ja'
    is_nsfw = True
    status = 'disabled'

    has_cf = True

    base_url = 'https://mangaraw.org'
    search_url = base_url + '/search'
    manga_url = base_url + '/{0}'
    chapter_url = base_url + '/{manga_slug}/{chapter_slug}'
    page_url = base_url + '/{0}/{1}/{2}'

    name_selector = '.infox h1'
    authors_selector = '.infox span:-soup-contains("Author")'
    genres_selector = '.infox span:-soup-contains("Genres") a'
    scanlators_selector = '.infox span:-soup-contains("Serialization")'
    status_selector = '.infox span:-soup-contains("Status")'
    synopsis_selector = '[itemprop="articleBody"]'

    def get_manga_chapters_data(self, soup):
        chapters = []
        for item in reversed(soup.select('.bixbox li')):
            a_element = item.find('a')

            slug = a_element.get('href').split('/')[-1]
            ignore = False
            for keyword in self.ignored_chapters_keywords:
                if keyword in slug:
                    ignore = True
                    break
            if ignore:
                continue

            chapters.append(dict(
                slug=slug,
                title=a_element.text.strip(),
                num=slug if is_number(slug) else None,
                date=convert_date_string(item.find('time').get('title').split()[0], format='%Y-%m-%d'),
            ))

        return chapters

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug=manga_slug, chapter_slug=chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            })
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )

        for option_element in soup.find_all('select', {'name': 'page'})[0].find_all('option'):
            data['pages'].append(dict(
                slug=option_element.get('value'),
                image=None,
            ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        headers = {
            'Referer': self.chapter_url.format(manga_slug=manga_slug, chapter_slug=chapter_slug),
        }
        r = self.session_get(self.page_url.format(manga_slug, chapter_slug, page['slug']), headers=headers)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')
        image_url = soup.select_one('.reader a img.picture').get('src')

        r = self.session_get(image_url, headers=headers)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=image_url.split('/')[-1],
        )

    def get_latest_updates(self, type):
        """
        Returns list of latest updates
        """
        return self.search('', type, orderby='latest')

    def get_most_populars(self, type):
        """
        Returns list of most popular manga
        """
        return self.search('', type, orderby='populars')

    @CompleteChallenge()
    def search(self, term, type, orderby=None):
        if orderby:
            data = dict(
                order='popular' if orderby == 'populars' else 'update',
                status='',
                type=type,
            )
        else:
            data = dict(
                s=term,
                author='',
                released='',
                status='',
                order='title',
                type=type,
            )

        r = self.session_get(self.search_url, params=data)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('.bsx > a'):
            type_ = a_element.select_one('.type').text.strip()
            if type_ == 'Novel':
                continue

            img_element = a_element.select_one('img')
            slug = a_element.get('href').split('/')[-1]

            results.append(dict(
                slug=slug,
                name=img_element.get('alt').strip(),
                cover=img_element.get('src'),
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara

#
# BEWARE: Wakascan server is disabled
# Replaced by an anim site since 01/17/2021
#


class Wakascan(Madara):
    id = 'wakascan'
    name = 'Wakascan'
    lang = 'fr'
    status = 'disabled'

    base_url = 'https://wakascan.com'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from bs4 import BeautifulSoup

from komikku.servers.multi.keyoapp import Keyoapp
from komikku.utils import get_buffer_mime_type


class Writerscans(Keyoapp):
    id = 'writerscans'
    name = 'Writer Scans'
    lang = 'en'

    base_url = 'https://writerscans.com'
    logo_url = 'https://wsrv.nl/?url=cdn.meowing.org/uploads/9fo4CQEukhQ&w=20'

    def get_most_populars(self):
        """
        Returns most popular manga
        """
        r = self.session_get(self.most_populars_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('div.grid div:-soup-contains("Trending") button > a'):
            cover = a_element.div.div.get('style').replace('background-image:url', '')[1:-1]

            results.append(dict(
                slug=a_element.get('href').split('/')[-2],
                name=a_element.get('title'),
                cover=cover,
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import json
import re

from bs4 import BeautifulSoup

from komikku.servers.multi.madara import Madara
from komikku.utils import get_buffer_mime_type


class Frdashscan(Madara):
    id = 'frdashscan'
    name = 'Fr-Scan'
    lang = 'fr'
    is_nsfw = True
    status = 'disabled'

    date_format = None

    base_url = 'https://fr-scan.com'
    chapter_url = base_url + '/manga/{0}/{1}/'  # don't support style param
    chapters_url = base_url + '/manga/{0}/ajax/chapters/'

    images_regex = r'var chapter_preloaded_images *= *(\[.*\]).*'

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        if script_element := soup.select_one('script#chapter_preloaded_images'):
            data = dict(
                pages=[],
            )

            script = script_element.string.strip()
            for line in script.split(';'):
                line = line.strip()
                if not line.startswith('var chapter_preloaded_images'):
                    continue

                line = line.replace('\\', '')
                if match := re.compile(self.images_regex).search(line):
                    for url in json.loads(match.group(1)):
                        data['pages'].append(dict(
                            slug=None,
                            image=url,
                        ))

                return data
        else:
            return Madara.get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import base64
import json
import re

from bs4 import BeautifulSoup
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import modes

from komikku.servers.multi.my_manga_reader_cms import MyMangaReaderCMS
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type


def md5(text):
    """
    Returns the MD5 digest of ``text`` as bytes

    ``text`` is a ``bytes`` instance
    """
    digest = hashes.Hash(hashes.MD5())
    digest.update(text)

    return digest.finalize()


def generate_key(passphrase, salt):
    """
    Generates the key from ``passphrase`` and ``salt`` as bytes

    ``passphrase`` is a ``bytes`` instance
    ``salt`` is a ``bytes`` instance
    """
    passphrase += salt
    md5arr = [0] * 3
    md5arr[0] = md5(passphrase)

    key = md5arr[0]
    for i in range(1, 3):
        md5arr[i] = md5(md5arr[i - 1] + passphrase)
        key += md5arr[i]

    return key[:32]


class Mangasin(MyMangaReaderCMS):
    id = 'mangasin'
    name = 'M440.in (Mangas.in)'
    lang = 'es'
    is_nsfw = True

    search_query_param = 'q'

    base_url = 'https://m440.in'
    logo_url = base_url + '/favicon.ico'
    search_url = base_url + '/search'
    most_populars_url = base_url + '/filterList?page=1&cat=&alpha=&sortBy=views&asc=false&author=&tag=&artist='
    manga_url = base_url + '/manga/{0}'
    chapter_url = base_url + '/manga/{0}/{1}'
    image_url = None  # Images URLs can't be computed with manga/chapter/image slugs
    cover_url = base_url + '/uploads/manga/{0}/cover/cover_250x350.jpg'

    def get_manga_chapters_data(self, soup):
        chapters_data = None
        chapters_re = r'{(?=.*\\"ct\\")(?=.*\\"iv\\")(?=.*\\"s\\").*?}'

        for script_element in soup.select('script'):
            script = script_element.string
            if script is None or not re.findall(chapters_re, script):
                continue

            line = re.findall(chapters_re, script)[0]
            cdata = json.loads(json.loads(f'"{line}"'))

            # Decrypt
            passphrase = b'X^Ib1O*HLVh%3W2t'  # in js/ads2.js, must be deobfuscated
            dct = base64.b64decode(cdata['ct'])
            iv = bytes.fromhex(cdata['iv'])
            salt = bytes.fromhex(cdata['s'])
            key = generate_key(passphrase, salt)

            cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
            decryptor = cipher.decryptor()
            chapters_data = decryptor.update(dct) + decryptor.finalize()

            unpadder = padding.PKCS7(128).unpadder()
            chapters_data = unpadder.update(chapters_data)
            chapters_data = chapters_data + unpadder.finalize()

            chapters_data = json.loads(json.loads(chapters_data.decode('utf-8')))
            break

        if chapters_data is None:
            return []

        data = []
        for chapter in reversed(chapters_data):
            data.append(dict(
                slug=chapter['slug'],
                title=f'Vol {chapter["volume"]} - #{chapter["number"]} {chapter["name"]}',
                num=chapter['number'],
                num_volume=chapter['volume'],
                date=convert_date_string(chapter['updated_at'].split()[0], format='%Y-%m-%d'),
            ))

        return data

    def get_latest_updates(self):
        """
        Returns list of latest updated manga
        """
        r = self.session_get(
            self.base_url,
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('.mangalist .manga-item h3 a:nth-child(3)'):
            slug = a_element.get('href').split('/')[-1]
            results.append(dict(
                name=a_element.text.strip(),
                slug=slug,
                cover=self.cover_url.format(slug),
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.heancms import HeanCMS


class Hivetoon(HeanCMS):
    id = 'hivetoon'
    name = 'Hive Toon'
    lang = 'en'

    base_url = 'https://hivetoons.org'
    logo_url = base_url + '/favicon-32x32.png'
    api_url = 'https://api.hivetoons.org/api'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from functools import wraps
import json

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number

LANGUAGES_CODES = dict(
    en='en',
    es='es',
)


def set_lang(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        server = args[0]

        if not server.csrf_token:
            r = server.session_get(server.base_url)

            soup = BeautifulSoup(r.text, 'lxml')
            server.csrf_token = soup.select_one('meta[name="csrf-token"]')['content']

        if not server.is_lang_set:
            server.session_post(
                server.api_set_language,
                data={
                    'language': LANGUAGES_CODES[server.lang],
                },
                headers={
                    'X-CSRF-TOKEN': server.csrf_token,
                }
            )
            server.is_lang_set = True

        return func(*args, **kwargs)

    return wrapper


class Mangapluscreators(Server):
    id = 'mangapluscreators'
    name = 'MANGA Plus Creators by SHUEISHA'
    lang = 'en'
    status = 'disabled'  # commercial platforms are no supported

    is_lang_set = False
    csrf_token = None

    base_url = 'https://mangaplus-creators.jp/'
    logo_url = base_url + '/favicon.ico'
    search_url = base_url + '/keywords/'
    popular_url = base_url + '/titles/popular/'
    latest_updates_url = base_url + '/titles/recent?t=episode'
    manga_url = base_url + '/titles/{0}'
    chapter_url = base_url + '/episodes/{0}'
    api_url = base_url + '/api'
    api_set_language = api_url + '/language/set_language'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = {'User-Agent': USER_AGENT}

    @set_lang
    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping chapter HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status='ongoing',
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.select_one('.title').text.strip()
        data['cover'] = soup.select_one('.cover img').get('data-src')

        for element in soup.select('.tag-genre'):
            data['genres'].append(element.text.strip())

        for element in soup.select('.name-group .name'):
            data['authors'].append(element.text.strip())

        data['synopsis'] = soup.select_one('.summary').text.strip()

        for element in soup.select('.mod-item-series'):
            title = element.select_one('.number').text.strip()
            num = title[1:] if title.startswith('#') else None
            date = element.select_one('.latest-update').text.strip()

            data['chapters'].append({
                'slug': element.get('href').split('/')[-1],
                'title': title,
                'num': num if is_number(num) else None,
                'date': convert_date_string(date, '%Y-%m-%d'),
            })

        return data

    @set_lang
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        if element := soup.select_one('div[react="viewer"]'):
            pages = json.loads(element.get('data-pages'))
        else:
            return None

        data = dict(
            pages=[],
        )
        for page in pages['pc']:
            data['pages'].append({
                'slug': None,
                'image': page['image_url'],
                'index': page['page_no'],
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0:03d}.{1}'.format(page['index'], mime_type.split('/')[1]),
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    @set_lang
    def get_latest_updates(self):
        """
        Returns latest updates
        """
        r = self.session_get(self.latest_updates_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.item-recent'):
            name = element.select_one('.title-area .title').text.strip()
            cover = element.select_one('.image img').get('src')
            slug = cover.split('/')[5]  # serie slug is available in cover URL only
            last_chapter = element.select_one('.issue').text.strip()

            results.append(dict(
                slug=slug,
                name=name,
                cover=cover,
                last_chapter=last_chapter,
            ))

        return results

    @set_lang
    def get_most_populars(self):
        """
        Returns popular
        """
        r = self.session_get(self.popular_url, params={
            'p': 'm',  # 30 days
            'l': LANGUAGES_CODES[self.lang],
        })
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.item-recent'):
            name = element.select_one('.title-area .title').text.strip()
            cover = element.select_one('.image img').get('src')
            slug = cover.split('/')[5]  # serie slug is available in cover URL only
            last_chapter = element.select_one('.issue').text.strip()

            results.append(dict(
                slug=slug,
                name=name,
                cover=cover,
                last_chapter=last_chapter,
            ))

        return results

    def search(self, term):
        r = self.session_get(self.search_url, params={'q': term, 'l': self.lang})
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.item-search'):
            name = element.select_one('.title-area .title').text.strip()
            cover = element.select_one('.image img').get('src')
            slug = cover.split('/')[5]  # serie slug is available in cover URL only
            last_chapter = element.select_one('.issue').text.strip()

            results.append(dict(
                slug=slug,
                name=name,
                cover=cover,
                last_chapter=last_chapter,
            ))

        return results


class Mangapluscreators_es(Mangapluscreators):
    id = 'mangapluscreators_es'
    lang = 'es'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.genkan import Genkan


class Thenonamesscans(Genkan):
    id = 'thenonamesscans'
    name = 'The Nonames Scans'
    lang = 'en'
    status = 'disabled'

    base_url = 'https://the-nonames.com'
    search_url = base_url + '/comics?query={0}'
    most_populars_url = base_url + '/home'
    manga_url = base_url + '/comics/{0}'
    chapter_url = base_url + '/comics/{0}/{1}'
    image_url = base_url + '{0}'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type


class Honeymanga(Server):
    id = 'honeymanga'
    name = 'Honey Manga'
    lang = 'uk'

    base_url = 'https://honey-manga.com.ua'
    logo_url = base_url + '/favicon.ico'
    manga_url = base_url + '/book/{0}'
    chapter_url = base_url + '/read/{0}/{1}'
    resource_url = 'https://hmvolumestorage.b-cdn.net/public-resources/{0}'

    api_base_url = 'https://{0}.api.honey-manga.com.ua'
    api_search_url = api_base_url.format('search') + '/v2/manga/pattern'
    api_list_url = api_base_url.format('data') + '/v2/manga/cursor-list'
    api_manga_url = api_base_url.format('data') + '/manga/{0}'
    api_chapters_url = api_base_url.format('data') + '/v2/chapter/cursor-list'
    api_chapter_url = api_base_url.format('data') + '/chapter/frames/{0}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data via API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(
            self.api_manga_url.format(initial_data['slug']),
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = initial_data.copy()
        data.update(dict(
            name=resp_data['title'],
            cover=self.resource_url.format(resp_data['posterId']) + '?optimizer=image&width=512',
            authors=[],
            scanlators=[],  # Use teamId? Would require an additional query
            genres=resp_data['genresAndTags'],
            status=None,
            synopsis=resp_data['description'],
            chapters=[],
            server_id=self.id,
        ))

        # Authors
        for author in resp_data['authors']:
            if author not in data['authors']:
                data['authors'].append(author)
        for artist in resp_data['artists']:
            if artist not in data['authors']:
                data['authors'].append(artist)

        # Status
        if resp_data['titleStatus'] in ('Онгоінг', 'Анонс'):
            data['status'] = 'ongoing'
        elif resp_data['titleStatus'] == 'Завершено':
            data['status'] = 'complete'
        elif resp_data['titleStatus'] in ('Покинуто', 'Призупинено'):
            data['status'] = 'suspended'

        # Chapters
        data['chapters'] = list(reversed(self.get_manga_chapters_data(data['slug'])))

        return data

    def get_manga_chapters_data(self, manga_slug, page=1):
        """
        Returns manga chapters list via API
        """
        r = self.session_post(
            self.api_chapters_url,
            json={
                'mangaId': manga_slug,
                'page': page,
                'pageSize': 1000,
                'sortOrder': 'DESC',
            },
            headers={
                'Content-Type': 'application/json',
                'Referer': f'{self.base_url}/',
                'X-Id-Token': 'undefined',
            }
        )
        if r.status_code != 200:
            return None

        chapters = []
        resp_data = r.json()
        for chapter in resp_data['data']:
            title = f'Том {chapter["volume"]} - Розділ {chapter["chapterNum"]}'
            if chapter['title'].lower().replace('-', '') not in ('', 'title'):
                # Not sure this restriction is exhaustive
                title += f'| {chapter["title"]}'

            chapters.append(dict(
                slug=chapter['id'],
                title=title,
                num=chapter['chapterNum'],
                num_volume=chapter['volume'],
                date=convert_date_string(chapter['lastUpdated'].split('T')[0], format='%Y-%m-%d'),
            ))

        if resp_data['cursorNext']:
            chapters += self.get_manga_chapters_data(manga_slug, resp_data['cursorNext']['page'])

        return chapters

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data via API

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.api_chapter_url.format(chapter_slug),
            headers={
                'Referer': self.chapter_url.format(chapter_slug, manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = dict(
            pages=[],
        )
        for index, resourceId in resp_data['resourceIds'].items():
            data['pages'].append(dict(
                slug=resourceId,
                index=int(index) + 1,
                image=None,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            self.resource_url.format(page['slug']),
            headers={
                'Referer': self.chapter_url.format(chapter_slug, manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0}.{1}'.format(page['index'], mime_type.split('/')[-1]),
        )

    def get_manga_list(self, orderby):
        r = self.session_post(
            self.api_list_url,
            json={
                'sort': {
                    'sortBy': orderby,
                    'sortOrder': 'DESC',
                },
                'page': 1,
                'pageSize': 30,
            },
            headers={
                'Origin': self.base_url,
                'Referer': self.base_url,
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        results = []
        for item in resp_data['data']:
            results.append(dict(
                slug=item['id'],
                name=item['title'],
                cover=self.resource_url.format(item['posterId']),
                nb_chapters=int(item['chapters']),
            ))

        return results

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        """
        Returns list of latest mangas via API
        """
        return self.get_manga_list('lastUpdated')

    def get_most_populars(self):
        """
        Returns list of most liked mangas via API
        """
        return self.get_manga_list('likes')

    def search(self, term):
        r = self.session_get(
            self.api_search_url,
            params=dict(
                query=term,
            ),
            headers={
                'Origin': self.base_url,
                'Referer': self.base_url,
            }
        )
        if r.status_code not in (200, 201):
            return None

        data = r.json()

        results = []
        for item in data:
            results.append(dict(
                slug=item['id'],
                name=item['title'],
                cover=self.resource_url.format(item['posterId']),
                nb_chapters=int(item['chapters']),
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from datetime import datetime
from functools import wraps
import re
import uuid

import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number

LANGUAGES_CODES = dict(
    de='deu',
    en='eng',
    es='esp',
    fr='fra',
    id='ind',
    pt_BR='ptb',
    ru='rus',
    th='tha',
    vi='vie',
)

LANGUAGES_CODES2 = dict(
    de='GERMAN',
    en='ENGLISH',
    es='SPANISH',
    fr='FRENCH',
    id='INDONESIAN',
    pt_BR='PORTUGUESE_BR',
    ru='RUSSIAN',
    th='THAI',
    vi='VIETNAMESE',
)

RE_ENCRYPTION_KEY = re.compile('.{1,2}')


def set_lang(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        server = args[0]
        if not server.is_lang_set:
            server.session_get(server.api_params_url, params=dict(lang=LANGUAGES_CODES[server.lang]))
            server.is_lang_set = True

        return func(*args, **kwargs)

    return wrapper


class Mangaplus(Server):
    id = 'mangaplus'
    name = 'MANGA Plus by SHUEISHA'
    lang = 'en'
    status = 'disabled'  # commercial platforms are no supported

    is_lang_set = False

    base_url = 'https://mangaplus.shueisha.co.jp'
    api_url = 'https://jumpg-webapi.tokyo-cdn.com/api'
    api_params_url = api_url + '/featured'
    api_search_url = api_url + '/title_list/allV2?format=json'
    # api_latest_updates_url = api_url + '/web/web_homeV4?lang={0}&format=json'
    api_latest_updates_url = api_url + '/home_v4?lang={0}&clang={0}&format=json'
    api_most_populars_url = api_url + '/title_list/rankingV2?lang={0}&type=hottest&clang={0}&format=json'
    api_manga_url = api_url + '/title_detailV3?title_id={0}&format=json'
    api_chapter_url = api_url + '/manga_viewer?chapter_id={0}&split=yes&img_quality=high&format=json'
    manga_url = base_url + '/titles/{0}'

    headers = {
        'User-Agent': USER_AGENT,
        'Origin': base_url,
        'Referer': f'{base_url}/',
        'SESSION-TOKEN': repr(uuid.uuid1()),
    }

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = self.headers

    @set_lang
    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.api_manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type not in ('application/json', 'text/plain'):
            return None

        resp_data = r.json()
        if resp_data.get('error'):
            return None

        resp_data = resp_data['success']['titleDetailView']

        data = initial_data.copy()
        data.update(dict(
            name=resp_data['title']['name'],
            authors=[resp_data['title']['author']],
            scanlators=['Shueisha'],
            genres=[],
            status='ongoing',
            synopsis=resp_data['overview'],
            chapters=[],
            server_id=self.id,
            cover=resp_data['title']['portraitImageUrl'],
        ))

        # Chapters
        for group in resp_data['chapterListGroup']:
            for key in ('firstChapterList', 'lastChapterList'):
                chapters = group.get(key)
                if chapters is None:
                    continue

                for chapter in chapters:
                    num = chapter['name'].lstrip('#')

                    data['chapters'].append(dict(
                        slug=str(chapter['chapterId']),
                        title='{0} - {1}'.format(chapter['name'], chapter['subTitle']),
                        num=num if is_number(num) else None,
                        date=datetime.fromtimestamp(chapter['startTimeStamp']).date(),
                    ))

        return data

    @set_lang
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data from API

        Currently, only pages are expected.
        """
        r = self.session_get(self.api_chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type not in ('application/json', 'text/plain'):
            return None

        resp_data = r.json()
        if resp_data.get('error'):
            return None

        data = dict(
            pages=[],
        )
        for page in resp_data['success']['mangaViewer']['pages']:
            if not page.get('mangaPage'):
                continue

            data['pages'].append(dict(
                slug=None,
                image=page['mangaPage']['imageUrl'],
                encryption_key=page['mangaPage'].get('encryptionKey'),
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        if page['encryption_key'] is not None:
            # Decrypt
            key_stream = [int(v, 16) for v in RE_ENCRYPTION_KEY.findall(page['encryption_key'])]
            block_size_in_bytes = len(key_stream)

            content = bytes([int(v) ^ key_stream[index % block_size_in_bytes] for index, v in enumerate(r.content)])
        else:
            content = r.content

        mime_type = get_buffer_mime_type(content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=content,
            mime_type=mime_type,
            name=page['image'].split('?')[0].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    @set_lang
    def get_latest_updates(self):
        """
        Returns latest updates
        """
        r = self.session_get(
            self.api_latest_updates_url.format(LANGUAGES_CODES[self.lang]),
            headers={
                'Referer': self.base_url + '/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type not in ('application/json', 'text/plain'):
            return None

        resp_data = r.json()
        if resp_data.get('error'):
            return None

        results = []
        for group in resp_data['success']['homeViewV3']['groups']:
            for title_group in group['titleGroups']:
                for title in title_group['titles']:
                    info = title['title']

                    if info.get('language', 'ENGLISH') != LANGUAGES_CODES2[self.lang]:
                        continue

                    results.append(dict(
                        slug=info['titleId'],
                        name=info['name'],
                        cover=info['portraitImageUrl'],
                        last_chapter=title['chapterName'],
                    ))

        return results

    @set_lang
    def get_most_populars(self):
        """
        Returns hottest manga list
        """
        r = self.session_get(self.api_most_populars_url.format(LANGUAGES_CODES[self.lang]))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type not in ('application/json', 'text/plain'):
            return None

        resp_data = r.json()
        if resp_data.get('error'):
            return None

        results = []
        for group in resp_data['success']['titleRankingViewV2']['rankedTitles']:
            for title in group['titles']:
                if title.get('language', 'ENGLISH') != LANGUAGES_CODES2[self.lang]:
                    continue

                results.append(dict(
                    slug=title['titleId'],
                    name=title['name'],
                    cover=title['portraitImageUrl'],
                ))

        return results

    @set_lang
    def search(self, term):
        r = self.session_get(self.api_search_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type not in ('application/json', 'text/plain'):
            return None

        resp_data = r.json()
        if resp_data.get('error'):
            return None

        results = []
        for group in resp_data['success']['allTitlesViewV2']['AllTitlesGroup']:
            for title in group['titles']:
                if title.get('language', 'ENGLISH') != LANGUAGES_CODES2[self.lang]:
                    continue

                if term.lower() not in title['name'].lower():
                    continue

                results.append(dict(
                    slug=title['titleId'],
                    name=title['name'],
                    cover=title['portraitImageUrl'],
                ))

        return results


class Mangaplus_de(Mangaplus):
    id = 'mangaplus_de'
    lang = 'de'


class Mangaplus_es(Mangaplus):
    id = 'mangaplus_es'
    lang = 'es'


class Mangaplus_fr(Mangaplus):
    id = 'mangaplus_fr'
    lang = 'fr'


class Mangaplus_id(Mangaplus):
    id = 'mangaplus_id'
    lang = 'id'


class Mangaplus_pt_br(Mangaplus):
    id = 'mangaplus_pt_br'
    lang = 'pt_BR'


class Mangaplus_ru(Mangaplus):
    id = 'mangaplus_ru'
    lang = 'ru'


class Mangaplus_th(Mangaplus):
    id = 'mangaplus_th'
    lang = 'th'


class Mangaplus_vi(Mangaplus):
    id = 'mangaplus_vi'
    lang = 'vi'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara2


class Webtoonhatti(Madara2):
    id = 'webtoonhatti'
    name = 'Webtoon Hatti'
    lang = 'tr'
    is_nsfw = True

    date_format = None
    series_name = 'webtoon'

    base_url = 'https://webtoonhatti.club'
    logo_url = base_url + '/wp-content/uploads/2024/03/cropped-iconlogo.png'

    def __init__(self):
        super().__init__()
        self.session.verify = False



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import datetime
from gettext import gettext as _
import json
from urllib.parse import unquote

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number


class Flamescans(Server):
    id = 'flamescans'
    name = 'Flame Comics (Flame Scans)'
    lang = 'en'

    base_url = 'https://flamecomics.xyz'
    logo_url = base_url + '/favicon.ico'
    search_url = base_url + '/browse'
    manga_url = base_url + '/series/{0}'
    chapter_url = base_url + '/series/{0}/{1}'
    cover_url = base_url + '/_next/image?url=https%3A%2F%2Fcdn.flamecomics.xyz%2Fuploads%2Fimages%2Fseries%2F{0}%2F{1}%3F%2F{2}&w=720&q=100'
    long_strip_genres = ['Manhua', 'Manhwa']

    filters = [
        {
            'key': 'types',
            'type': 'select',
            'name': _('Types'),
            'description': _('Filter by Types'),
            'value_type': 'multiple',
            'options': [
                {'key': 'Manga', 'name': _('Manga'), 'default': False},
                {'key': 'Manhwa', 'name': _('Manhwa'), 'default': False},
                {'key': 'Manhua', 'name': _('Manhua'), 'default': False},
            ],
        },
        {
            'key': 'statuses',
            'type': 'select',
            'name': _('Statuses'),
            'description': _('Filter by Statuses'),
            'value_type': 'multiple',
            'options': [
                {'key': 'Ongoing', 'name': _('Ongoing'), 'default': False},
                {'key': 'Completed', 'name': _('Completed'), 'default': False},
                {'key': 'Hiatus', 'name': _('Hiatus'), 'default': False},
                {'key': 'Dropped', 'name': _('Dropped'), 'default': False},
                {'key': 'Cancelled', 'name': _('Canceled'), 'default': False},
            ]
        },
    ]

    ignored_images_keywords = ['read_on_flame']

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        soup = BeautifulSoup(r.text, 'lxml')

        found = False
        for script_element in soup.select('script'):
            script = script_element.string
            if not script or script_element.get('id') != '__NEXT_DATA__':
                continue

            try:
                serie_data = json.loads(script)['props']['pageProps']['series']
            except Exception:
                break

            found = True
            data['name'] = serie_data['title']
            data['cover'] = self.cover_url.format(data['slug'], serie_data['cover'], serie_data['last_edit'])

            # Details
            if serie_data.get('author'):
                for author in serie_data['author']:
                    data['authors'].append(author)
            if serie_data.get('artist'):
                for artist in serie_data['artist']:
                    if artist in data['authors']:
                        continue
                    data['authors'].append(artist)

            if serie_data.get('tags'):
                data['genres'] = serie_data['tags']
            if serie_data.get('type'):
                data['genres'].append(serie_data['type'])

            if serie_data['status'] == 'Ongoing':
                data['status'] = 'ongoing'
            elif serie_data['status'] == 'Completed':
                data['status'] = 'complete'
            elif serie_data['status'] in ('Dropped', 'Cancelled'):
                data['status'] = 'suspended'
            elif serie_data['status'] == 'Hiatus':
                data['status'] = 'hiatus'
            else:
                data['genres'].append(serie_data['status'])

            if serie_data.get('description'):
                data['synopsis'] = BeautifulSoup(serie_data['description'], 'lxml').text

            # Chapters
            chapters_data = json.loads(script)['props']['pageProps']['chapters']
            for chapter in reversed(chapters_data):
                if is_number(chapter['chapter']):
                    if float(chapter['chapter']) == int(float(chapter['chapter'])):
                        num = int(float(chapter['chapter']))
                    else:
                        num = float(chapter['chapter'])
                    title = f'Chapter {num}'
                else:
                    title = chapter['chapter']

                data['chapters'].append(dict(
                    slug=chapter['token'],
                    title=title,
                    num=num if is_number(num) else None,
                    date=datetime.date.fromtimestamp(chapter['release_date']),
                ))

        if not found:
            return None

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for element in soup.select('img[my="none"]'):
            ignore = False
            image = element.get('src')
            for keyword in self.ignored_images_keywords:
                if keyword in image:
                    ignore = True
                    break

            if ignore:
                continue

            data['pages'].append(dict(
                slug=None,
                image=image,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        name = unquote(page['image'].split('url=')[-1]).split('?')[0].split('/')[-1]

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=name,
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, orderby=None, types=None, statuses=None):
        if term:
            r = self.session_get(self.search_url)
        else:
            r = self.session_get(self.base_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        found = False
        results = []
        for script_element in soup.select('script'):
            script = script_element.string
            if not script or script_element.get('id') != '__NEXT_DATA__':
                continue

            try:
                data = json.loads(script)
            except Exception:
                break

            found = True
            if term:
                series = data['props']['pageProps']['series']
            elif orderby == 'popular':
                series = data['props']['pageProps']['popularEntries']['blocks'][0]['series']
            elif orderby == 'latest':
                series = data['props']['pageProps']['latestEntries']['blocks'][0]['series']

            for serie in series:
                if term and term.lower() not in serie['title'].lower():
                    continue

                if types and serie['type'] not in types:
                    continue

                if statuses and serie['status'] not in statuses:
                    continue

                results.append({
                    'slug': serie['series_id'],
                    'name': serie['title'],
                    'cover': self.cover_url.format(serie['series_id'], serie['cover'], serie['last_edit']),
                })

        if not found:
            return None

        return results

    def get_latest_updates(self, types=None, statuses=None):
        return self.get_manga_list(orderby='latest', statuses=statuses, types=types)

    def get_most_populars(self, types=None, statuses=None):
        return self.get_manga_list(orderby='popular', statuses=statuses, types=types)

    def search(self, term='', types=None, statuses=None):
        return self.get_manga_list(term=term, statuses=statuses, types=types)



# SPDX-FileCopyrightText: 2023-2024 Pierre-Emmanuel Devin
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Pierre-Emmanuel Devin <pierreemmanuel.devin@posteo.net>
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import logging
import time
from urllib.parse import unquote

from bs4 import BeautifulSoup
import requests

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import parse_nextjs_hydration
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_response_elapsed
from komikku.utils import is_number

logger = logging.getLogger(__name__)


class Littlexgarden(Server):
    id = 'littlexgarden'
    name = 'Punk Records (Little Garden)'
    lang = 'fr'

    base_url = 'https://punkrecordz.com'
    logo_url = base_url + '/favicon.ico'
    manga_list_url = base_url + '/mangas'
    manga_url = base_url + '/mangas/{0}'
    chapter_url = base_url + '/mangas/{0}/{1}?display=vertical'

    api_base_url = 'https://api.punkrecordz.com'
    api_url = api_base_url + '/graphql'
    image_url = api_base_url + '/images/webp/{0}.webp'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data via GraphQL API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        # Chapters
        def get_page(cursor, limit):
            r = self.session_post(
                self.api_url,
                json={
                    'operationName': 'chapters',
                    'query': 'query chapters($slug: String, $number: Float, $limit: Int = 12, $skip: Int = 0, $order: Float = -1) {\n  chapters(\n    limit: $limit\n    skip: $skip\n    where: {number: $number, deleted: false, published: true, manga: {slug: $slug, published: true, deleted: false}}\n    order: [{field: \"number\", order: $order}]\n  ) {\n    id\n    likes\n    number\n    thumb\n  }\n}',
                    'variables': {
                        'limit': limit,
                        'order': 1,
                        'skip': cursor,
                        'slug': initial_data['slug'],
                    },
                },
                headers={
                    'Content-Type': 'application/json',
                    'Referer': f'{self.base_url}/',
                }
            )
            if r.status_code != 200:
                return None, None

            chunk = r.json()['data']['chapters']
            count = len(chunk)
            next_cursor = cursor + count if count > 0 else None

            return chunk, next_cursor, get_response_elapsed(r)

        chapters = []
        cover = None
        cursor = 0
        delay = None
        while cursor is not None:
            if delay:
                time.sleep(delay)

            chunk, cursor, rtime = get_page(cursor, 500)
            for chapter in chunk:
                num = chapter['number']

                chapters.append(dict(
                    slug=str(num),  # slug nust be a string
                    title=f'Chapitre {num}',
                    num=num if is_number(num) else None,
                    date=None,
                ))

                if cover is None:
                    # Use 1st image of 1st chapter as cover
                    cover = self.image_url.format(chapter['thumb'])

            delay = min(rtime * 2, DOWNLOAD_MAX_DELAY) if rtime else None

        data = initial_data.copy()
        data.update(dict(
            authors=[],  # not available
            scanlators=[],  # not available
            genres=[],  # not available
            status=None,  # not available
            synopsis=None,  # not available
            chapters=chapters,
            server_id=self.id,
            cover=cover,
        ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = {
            'pages': [],
        }
        for index, element in enumerate(soup.select('img[data-nimg="1"]'), start=1):
            data['pages'].append(dict(
                slug=unquote(element.get('src')).split('/')[-1].split('.webp')[0],
                image=None,
                index=index,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(self.image_url.format(page['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0:03d}.{1}'.format(page['index'], mime_type.split('/')[1]),
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute url
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        """
        Returns latest updated manga
        """
        r = self.session_get(self.base_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        slugs = []
        for a_element in soup.select('section:-soup-contains("Les derniers scans") a'):
            slug = a_element.get('href').split('/')[1]
            if slug in slugs:
                continue

            cover = None
            if cover_element := a_element.select_one('div > div > div:last-child > div'):
                for style in cover_element.get('style').split(';'):
                    if not style.startswith('background-image'):
                        continue
                    cover = style[21:-1]

            results.append(dict(
                name=a_element.select_one('h4').text.strip(),
                slug=slug,
                cover=cover,
                last_chapter=a_element.select_one('p.chakra-text').text.strip(),
            ))
            slugs.append(slug)

        return results

    def get_manga_list(self):
        """
        Returns all manga
        """
        r = self.session_get(self.manga_list_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        info = parse_nextjs_hydration(soup, 'slug')
        if info is None:
            return None

        results = []
        for item in info[3]['children'][3]['data']:
            results.append(dict(
                name=item['name'],
                slug=item['slug'],
                cover=self.image_url.format(item['thumb']),
            ))

        return results

    def get_most_populars(self):
        return self.get_manga_list()

    def search(self, term):
        mangas = self.get_manga_list()
        if mangas is None:
            return None

        results = []
        for manga in mangas:
            if term.lower() in manga['name'].lower():
                results.append(manga)

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import shutil

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

from komikku.models import Settings
from komikku.reader.controls import Controls
from komikku.reader.pager import Pager
from komikku.reader.pager.webtoon import WebtoonPager
from komikku.utils import get_file_mime_type


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/reader.ui')
class ReaderPage(Adw.NavigationPage):
    __gtype_name__ = 'ReaderPage'

    headerbar_revealer = Gtk.Template.Child('headerbar_revealer')
    back_button = Gtk.Template.Child('back_button')
    title = Gtk.Template.Child('title')
    fullscreen_button = Gtk.Template.Child('fullscreen_button')
    menu_button = Gtk.Template.Child('menu_button')

    overlay = Gtk.Template.Child('reader_overlay')

    manga = None
    init_chapter = None
    chapters_consulted = None
    pager = None

    def __init__(self, window):
        Adw.NavigationPage.__init__(self)

        self.window = window
        self.builder = window.builder
        self.builder.add_from_resource('/info/febvre/Komikku/ui/menu/reader.xml')

        self.connect('hidden', self.on_hidden)
        self.connect('shown', self.on_shown)
        self.window.connect('notify::fullscreened', self.on_fullscreen_state_changed)
        self.window.controller_key.connect('key-pressed', self.on_key_pressed)

        # Double maximum distance allowed between two clicks (default 5)
        # Allows to zoom more easily (double tap) on touch screen
        Gtk.Settings.get_default().set_property('gtk-double-click-distance', 10)

        # Header bar
        self.back_button.connect('clicked', self.on_back_button_clicked)
        self.fullscreen_button.connect('clicked', self.window.toggle_fullscreen, None)
        self.menu_button.set_menu_model(self.builder.get_object('menu-reader'))
        # Focus is lost after showing popover submenu (bug?)
        self.menu_button.get_popover().connect('closed', lambda _popover: self.menu_button.grab_focus())

        # Page numbering
        self.page_numbering_defined = False
        self.page_numbering_label = Gtk.Label(halign=Gtk.Align.CENTER, can_focus=False, can_target=False)
        self.page_numbering_label.add_css_class('reader-page-number-indicator-label')
        self.page_numbering_label.set_valign(Gtk.Align.END)
        self.overlay.add_overlay(self.page_numbering_label)

        # Controls
        self.controls = Controls(self)

        # Settings dialog
        self.settings_dialog = SettingsDialog(self)

        self.window.navigationview.add(self)

    @property
    def background_color(self):
        return self.manga.background_color or Settings.get_default().background_color

    @property
    def borders_crop(self):
        if self.reading_mode == 'webtoon':
            # Borders crop is not an option in Webtoon reading mode
            # Ignore settings and return False
            return False

        if self.manga.borders_crop in (0, 1):
            return bool(self.manga.borders_crop)

        return Settings.get_default().borders_crop

    @property
    def landscape_zoom(self):
        if self.reading_mode == 'webtoon':
            # Landscape zoom is not an option in Webtoon reading mode
            # Ignore settings and return False
            return False

        if self.manga.landscape_zoom in (0, 1):
            return bool(self.manga.landscape_zoom)

        return Settings.get_default().landscape_zoom

    @property
    def page_numbering(self):
        if self.manga.page_numbering in (0, 1):
            return bool(self.manga.page_numbering)

        return Settings.get_default().page_numbering

    @property
    def reading_mode(self):
        return self.manga.reading_mode or Settings.get_default().reading_mode

    @property
    def scaling(self):
        if self.reading_mode == 'webtoon':
            # Scaling is not an option in Webtoon reading mode
            # Ignore settings and return 'width'
            return 'width'

        return self.manga.scaling or Settings.get_default().scaling

    @property
    def scaling_filter(self):
        return self.manga.scaling_filter or Settings.get_default().scaling_filter

    @property
    def size(self):
        width = self.window.get_width()
        height = self.window.get_height()

        if self.headerbar_revealer.get_child_revealed():
            height -= self.get_child().get_top_bar_height()

        return (width, height)

    def add_accelerators(self):
        self.window.application.set_accels_for_action('app.reader.save-page', ['<Primary>s'])

    def add_actions(self):
        # Reading mode
        variant = GLib.Variant.new_string('right-to-left')
        self.reading_mode_action = Gio.SimpleAction.new_stateful('reader.reading-mode', variant.get_type(), variant)
        self.reading_mode_action.connect('activate', self.on_reading_mode_changed)
        self.window.application.add_action(self.reading_mode_action)

        # Scaling
        variant = GLib.Variant.new_string('screen')
        self.scaling_action = Gio.SimpleAction.new_stateful('reader.scaling', variant.get_type(), variant)
        self.scaling_action.connect('activate', self.on_scaling_changed)
        self.window.application.add_action(self.scaling_action)

        # Scaling filter
        variant = GLib.Variant.new_string('linear')
        self.scaling_filter_action = Gio.SimpleAction.new_stateful('reader.scaling-filter', variant.get_type(), variant)
        self.scaling_filter_action.connect('activate', self.on_scaling_filter_changed)
        self.window.application.add_action(self.scaling_filter_action)

        # Landscape Pages Zoom
        self.landscape_zoom_action = Gio.SimpleAction.new_stateful('reader.landscape-zoom', None, GLib.Variant('b', False))
        self.landscape_zoom_action.connect('change-state', self.on_landscape_zoom_changed)
        self.window.application.add_action(self.landscape_zoom_action)

        # Background color
        variant = GLib.Variant.new_string('white')
        self.background_color_action = Gio.SimpleAction.new_stateful('reader.background-color', variant.get_type(), variant)
        self.background_color_action.connect('activate', self.on_background_color_changed)
        self.window.application.add_action(self.background_color_action)

        # Borders crop
        self.borders_crop_action = Gio.SimpleAction.new_stateful('reader.borders-crop', None, GLib.Variant('b', False))
        self.borders_crop_action.connect('change-state', self.on_borders_crop_changed)
        self.window.application.add_action(self.borders_crop_action)

        # Page numbering
        self.page_numbering_action = Gio.SimpleAction.new_stateful('reader.page-numbering', None, GLib.Variant('b', False))
        self.page_numbering_action.connect('change-state', self.on_page_numbering_changed)
        self.window.application.add_action(self.page_numbering_action)

        # Save page
        self.save_page_action = Gio.SimpleAction.new('reader.save-page', None)
        self.save_page_action.connect('activate', self.save_page)
        self.window.application.add_action(self.save_page_action)

        # Open settings dialog
        self.open_settings_dialog_action = Gio.SimpleAction.new('reader.open-settings-dialog', None)
        self.open_settings_dialog_action.connect('activate', self.open_settings_dialog)
        self.window.application.add_action(self.open_settings_dialog_action)

    def init(self, manga, chapter):
        self.manga = manga
        self.init_chapter = chapter

        # Reset list of chapters consulted
        self.chapters_consulted = set()

        # Init settings
        self.set_action_reading_mode()
        self.set_action_scaling()
        self.set_action_scaling_filter()
        self.set_action_landscape_zoom()
        self.set_action_borders_crop()
        self.set_action_page_numbering()

        if Settings.get_default().fullscreen:
            self.window.fullscreen()

        self.back_button.set_tooltip_text(self.manga.name)

        self.show()

    def init_pager(self, chapter):
        if self.pager:
            self.pager.dispose()

        if self.reading_mode == 'webtoon':
            self.pager = WebtoonPager(self)
        else:
            self.pager = Pager(self)
            self.set_orientation()

        self.set_action_background_color()

        self.overlay.set_child(self.pager)

        self.pager.init(chapter)

    def on_back_button_clicked(self, _btn):
        self.window.navigationview.pop()

    def on_background_color_changed(self, _action, variant):
        value = variant.get_string()
        if value == self.background_color:
            return

        self.manga.update(dict(
            background_color=value if value != Settings.get_default().background_color else None,
        ))
        self.set_action_background_color()

    def on_borders_crop_changed(self, _action, variant):
        value = variant.get_boolean()
        if value == self.borders_crop:
            return

        self.manga.update(dict(
            borders_crop=value if value != Settings.get_default().borders_crop else None,
        ))
        self.set_action_borders_crop()

        self.pager.crop_pages_borders()

    def on_fullscreen_state_changed(self, _window, gparam):
        if self.window.is_fullscreen():
            self.headerbar_revealer.set_reveal_child(False)
            self.fullscreen_button.set_icon_name('view-restore-symbolic')
        else:
            self.headerbar_revealer.set_reveal_child(True)
            self.fullscreen_button.set_icon_name('view-fullscreen-symbolic')

    def on_hidden(self, _page):
        if self.window.previous_page == self.props.tag:
            return

        if self.pager:
            self.pager.dispose()
            self.pager = None

        self.controls.hide()
        self.page_numbering_label.set_visible(False)
        self.window.unfullscreen()

        # Sync Card page
        if self.window.card in self.window.navigationview.get_navigation_stack():
            if self.chapters_consulted:
                # Refresh to update all previously chapters consulted (last page read may have changed also)
                # and update info like disk usage
                self.window.card.refresh(unread_chapters=True, info=True, chapters=self.chapters_consulted)

        # Sync History page
        if self.window.history in self.window.navigationview.get_navigation_stack():
            self.window.history.populate()

        # Sync Library page (root)
        if self.manga.in_library:
            self.window.library.update_thumbnail(self.manga)
            self.window.library.flowbox.invalidate_sort()

    def on_key_pressed(self, _controller, keyval, _keycode, state):
        if self.window.page != self.props.tag:
            return Gdk.EVENT_PROPAGATE

        # Allow to navigate back with <Esc> key and <Alt+Left>
        modifiers = state & Gtk.accelerator_get_default_mod_mask()
        if keyval == Gdk.KEY_Escape or (modifiers == Gdk.ModifierType.ALT_MASK and keyval in (Gdk.KEY_Left, Gdk.KEY_KP_Left)):
            self.window.navigationview.pop()
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def on_landscape_zoom_changed(self, _action, variant):
        value = variant.get_boolean()
        if value == self.landscape_zoom:
            return

        self.manga.update(dict(
            landscape_zoom=value if value != Settings.get_default().landscape_zoom else None,
        ))
        self.set_action_landscape_zoom()

        self.pager.rescale_pages()

    def on_page_numbering_changed(self, _action, variant):
        value = not variant.get_boolean()
        if value == self.page_numbering:
            return

        self.manga.update(dict(
            page_numbering=value if value != Settings.get_default().page_numbering else None,
        ))
        self.set_action_page_numbering()

        if value and self.page_numbering_defined and not self.controls.is_visible:
            self.page_numbering_label.set_visible(True)
        else:
            self.page_numbering_label.set_visible(False)

    def on_reading_mode_changed(self, _action, variant):
        value = variant.get_string()
        if value == self.reading_mode:
            return

        prior_reading_mode = self.reading_mode

        self.manga.update(dict(
            reading_mode=value if value != Settings.get_default().reading_mode else None,
        ))
        self.set_action_reading_mode()

        if value == 'webtoon' or prior_reading_mode == 'webtoon':
            self.init_pager(self.pager.current_page.chapter)
        else:
            if value == 'right-to-left' or prior_reading_mode == 'right-to-left':
                self.pager.reverse_pages()
            self.set_orientation()

    def on_scaling_changed(self, _action, variant):
        value = variant.get_string()
        if value == self.scaling:
            return

        self.manga.update(dict(
            scaling=value if value != Settings.get_default().scaling else None,
        ))
        self.set_action_scaling()

        self.pager.rescale_pages()

    def on_scaling_filter_changed(self, _action, variant):
        value = variant.get_string()
        if value == self.scaling:
            return

        self.manga.update(dict(
            scaling_filter=value if value != Settings.get_default().scaling_filter else None,
        ))
        self.set_action_scaling_filter()

        self.pager.rescale_pages()

    def on_shown(self, _page):
        def do_init_pager():
            if self.window.last_navigation_action != 'push':
                return

            # Wait page is shown (transition is ended) to init pager
            # Operation is resource intensive and could disrupt page transition
            self.init_pager(self.init_chapter)

        if not Gtk.Settings.get_default().get_property('gtk-enable-animations'):
            # When animations are disabled, popped/pushed events are sent after `shown` event (bug?)
            # Use idle_add to be sure that last `popped` or `pushed` event has been received
            GLib.idle_add(do_init_pager)
        else:
            do_init_pager()

    def open_settings_dialog(self, _action, _gparam):
        self.settings_dialog.present(self.window)

    def save_page(self, _action, _gparam):
        if self.window.page != self.props.tag:
            return

        page = self.pager.current_page
        if not page.image or page.error:
            return

        def do_save(dest_path):
            try:
                shutil.copy(page.path, dest_path)
            except Exception as error:
                self.window.add_notification(str(error))
            else:
                self.window.add_notification(_('Page successfully saved'))

        def on_ready(dialog, result):
            try:
                gfile = dialog.save_finish(result)
            except GLib.GError:
                # Cancel
                gfile = None

            if gfile:
                do_save(gfile.get_path())

        extension = get_file_mime_type(page.path).split('/')[-1]
        filename = f'{self.manga.name}_{page.chapter.title}_{str(page.index + 1)}.{extension}'
        xdg_pictures_dir = GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_PICTURES)

        dialog = Gtk.FileDialog(modal=True)
        dialog.set_initial_name(filename)
        if xdg_pictures_dir is not None:
            dialog.set_initial_folder(Gio.File.new_for_path(xdg_pictures_dir))

        dialog.save(self.window, None, on_ready)

    def set_action_background_color(self):
        self.background_color_action.set_state(GLib.Variant('s', self.background_color))

        if self.background_color == 'white':
            self.pager.set_css_classes(['background-white'])
        elif self.background_color == 'gray':
            self.pager.set_css_classes(['background-gray'])
        elif self.background_color == 'black':
            self.pager.set_css_classes(['background-black'])
        else:
            # System style
            self.pager.set_css_classes([])

    def set_action_borders_crop(self):
        self.borders_crop_action.set_state(GLib.Variant('b', self.borders_crop))

    def set_action_landscape_zoom(self):
        self.landscape_zoom_action.set_state(GLib.Variant('b', self.landscape_zoom))

    def set_action_page_numbering(self):
        self.page_numbering_action.set_state(GLib.Variant('b', not self.page_numbering))

    def set_action_reading_mode(self):
        self.reading_mode_action.set_state(GLib.Variant('s', self.reading_mode))

        # Scaling action is enabled in RTL/LTR/Vertical reading modes only
        self.scaling_action.set_enabled(self.reading_mode != 'webtoon')
        # Landscape pages zoom is enabled in RTL/LTR/Vertical reading modes only and when scaling is 'screen'
        self.landscape_zoom_action.set_enabled(self.reading_mode != 'webtoon' and self.scaling == 'screen')
        # Borders crop is enabled in RTL/LTR/Vertical reading modes only
        self.borders_crop_action.set_enabled(self.reading_mode != 'webtoon')

        # Additionally, direction of page slider in controls must be updated
        self.controls.set_scale_direction(inverted=self.reading_mode == 'right-to-left')

    def set_action_scaling(self, scaling=None):
        self.scaling_action.set_state(GLib.Variant('s', scaling or self.scaling))

        # Landscape pages zoom is enabled in RTL/LTR/Vertical reading modes only and when scaling is 'screen'
        self.landscape_zoom_action.set_enabled(self.reading_mode != 'webtoon' and self.scaling == 'screen')

    def set_action_scaling_filter(self, scaling_filter=None):
        self.scaling_filter_action.set_state(GLib.Variant('s', scaling_filter or self.scaling_filter))

    def set_orientation(self):
        if self.reading_mode in ('right-to-left', 'left-to-right'):
            orientation = Gtk.Orientation.HORIZONTAL
        else:
            orientation = Gtk.Orientation.VERTICAL

        self.pager.set_orientation(orientation)

    def show(self):
        self.window.navigationview.push(self)

    def toggle_controls(self, visible=None):
        if visible is None:
            visible = not self.controls.is_visible

        if visible:
            self.controls.show()
            self.page_numbering_label.set_visible(False)
        else:
            self.controls.hide()
            if self.page_numbering and self.page_numbering_defined:
                self.page_numbering_label.set_visible(True)

    def update_page_numbering(self, number=None, total=None):
        if number and total:
            self.page_numbering_label.set_text('{0}/{1}'.format(number, total))
            self.page_numbering_defined = True

            if self.page_numbering and not self.controls.is_visible:
                self.page_numbering_label.set_visible(True)
            else:
                self.page_numbering_label.set_visible(False)
        else:
            self.page_numbering_defined = False
            self.page_numbering_label.set_visible(False)

    def update_title(self, chapter):
        # Set title & subtitle (headerbar)
        self.title.set_title(chapter.manga.name)
        subtitle = chapter.title
        if chapter.manga.name in subtitle:
            subtitle = subtitle.replace(chapter.manga.name, '').strip()
        self.title.set_subtitle(subtitle)


class SettingsDialog(Adw.PreferencesDialog):
    def __init__(self, page):
        super().__init__()
        self.page = page
        self.settings = Settings.get_default()

        self.set_title(_('Settings'))
        self.set_presentation_mode(Adw.DialogPresentationMode.BOTTOM_SHEET)

        self.css_provider = Gtk.CssProvider.new()
        Gtk.StyleContext.add_provider_for_display(Gdk.Display.get_default(), self.css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        page = Adw.PreferencesPage(title=_('Filters'))
        self.add(page)

        group = Adw.PreferencesGroup(title=_('Filters'), separate_rows=True)

        filters = {
            'brightness': {
                'title': _('Brightness'),
                'subtitle': _('Make pages brighter or darker'),
                'min': 0,
                'max': 200,
                'step': 1,
                'default': 100,
            },
            'contrast': {
                'title': _('Contrast'),
                'subtitle': _('Increase or decrease the contrast of pages'),
                'min': 0,
                'max': 200,
                'step': 1,
                'default': 100,
            },
            'grayscale': {
                'title': _('Grayscale'),
                'subtitle': _('Convert pages to grayscale'),
                'min': 0,
                'max': 100,
                'step': 1,
                'default': 0,
            },
            'sepia': {
                'title': _('Sepia'),
                'subtitle': _('Give a more yellow/brown appearance to pages'),
                'min': 0,
                'max': 100,
                'step': 1,
                'default': 0,
            },
            'saturate': {
                'title': _('Saturation'),
                'subtitle': _('Super-saturate or desaturate pages'),
                'min': 0,
                'max': 400,
                'step': 1,
                'default': 100,
            },
        }

        for name, data in filters.items():
            erow = Adw.ExpanderRow()
            erow.set_title(data['title'])
            erow.set_subtitle(data['subtitle'])
            erow.set_enable_expansion(self.settings.page_filters.get(f'{name}-state'))
            erow.set_show_enable_switch(True)
            erow.connect('notify::enable-expansion', self.toggle_filter_state, name)

            row = Adw.SpinRow.new_with_range(data['min'], data['max'], data['step'])
            row.set_title(data['title'])
            row.set_value(self.settings.page_filters.get(name, data['default']))
            row.connect('notify::value', self.set_filter_value, name)

            erow.add_row(row)
            group.add(erow)

        page.add(group)

        self.apply_filters()

    def apply_filters(self):
        funcs = []
        for name, value in self.settings.page_filters.items():
            if name.endswith('-state'):
                # Ignore state key
                continue
            if not self.settings.page_filters.get(f'{name}-state'):
                # Filter is off
                continue

            funcs.append(f'{name}({value}%)')

        if funcs:
            self.css_provider.load_from_string(f'.page-filters {{filter: {" ".join(funcs)};}}')  # noqa
            self.page.overlay.add_css_class('page-filters')
        else:
            self.css_provider.load_from_string('')
            self.page.overlay.remove_css_class('page-filters')

    def set_filter_value(self, row, _gparam, name):
        filters = self.settings.page_filters
        filters[name] = row.get_value()
        self.settings.page_filters = filters

        self.apply_filters()

    def toggle_filter_state(self, row, _gparam, name):
        filters = self.settings.page_filters
        filters[f'{name}-state'] = row.get_enable_expansion()
        self.settings.page_filters = filters

        self.apply_filters()



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import datetime
from functools import wraps
import glob
import importlib
import inspect
from io import BytesIO
import itertools
import json
import logging
import math
from operator import itemgetter
import os
from pkgutil import iter_modules
import re
import struct
import sys
import time

from bs4 import BeautifulSoup
from bs4 import NavigableString
import dateparser
import emoji
from gi.repository import Gio
from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont
import requests

from komikku.consts import USER_AGENT
from komikku.servers.loader import ServerFinder
from komikku.utils import get_cached_logos_dir

logger = logging.getLogger(__name__)


class TextImage:
    def __init__(self, text, width=1500, height=2126, bg_color='#fff', fg_color='#000', font_size=64, format='webp'):
        self.format = format
        self.image = Image.new('RGB', (width, height), bg_color)

        if text is None:
            text = ''

        rfont = Gio.resources_lookup_data('/info/febvre/Komikku/text-image.otf', Gio.ResourceLookupFlags.NONE)
        font = ImageFont.truetype(BytesIO(rfont.get_data()), font_size)

        draw = ImageDraw.Draw(self.image)
        if '\n' in text:
            left, top, right, bottom = draw.multiline_textbbox((0, 0), text, font, font_size=font_size)
        else:
            left, top, right, bottom = draw.textbbox((0, 0), text, font, font_size=font_size)

        text_width = right - left
        text_height = bottom - top
        text_coord = ((width - text_width) // 2, (height - text_height) // 2)

        draw.multiline_text(text_coord, text, fill=fg_color, font=font, align='center')
        del draw

    @property
    def content(self):
        buf = BytesIO()
        self.image.save(buf, self.format.upper())

        return buf.getvalue()

    @property
    def mime_type(self):
        return f'image/{self.format}'


def convert_date_string(date_string, format=None, languages=None):
    """
    Convert a date string into a date object

    :param date_string: A string representing date in a recognizably valid format
    :type date_string: str

    :param format: A format string using directives as given `here <https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior>`_
    :type format: str

    :param languages: A list of language codes, e.g. ['en', 'es', 'pt_BR']
    :type languages: list

    :return: A date object representing parsed date string if successful, `None` otherwise
    :rtype: datetime.date
    """

    # Check if languages are supported by dateparser
    # And detect whether a language code should be treated as a locale code
    if languages:
        language_locale = dateparser.data.language_locale_dict
        supported_languages = set()
        supported_locales = set()

        for code in languages:
            # Some codes should not be treated as locales
            if code.startswith(('zh_',)):
                code = code.replace('_', '-', 1)

            if '_' in code:
                lang, country = code.split('_')
            else:
                lang, country = code, None

            if lang not in language_locale:
                # Not supported
                continue

            if country and f'{lang}-{country}' in language_locale[lang]:
                # Code is a locale code
                supported_locales.add(f'{lang}-{country}')

            supported_languages.add(lang)

        languages = list(supported_languages)
        locales = list(supported_locales)
    else:
        locales = None

    if format is not None:
        try:
            d = datetime.datetime.strptime(date_string, format)
        except Exception:
            d = dateparser.parse(date_string, languages=languages, locales=locales)
    else:
        d = dateparser.parse(date_string, languages=languages, locales=locales)

    return d.date() if d else None


# https://github.com/italomaia/mangarock.py/blob/master/mangarock/mri_to_webp.py
def convert_mri_data_to_webp_buffer(data):
    size_list = [0] * 4
    size = len(data)
    header_size = size + 7

    # little endian byte representation
    # zeros to the right don't change the value
    for i, byte in enumerate(struct.pack('<I', header_size)):
        size_list[i] = byte

    buffer = [
        82,  # R
        73,  # I
        70,  # F
        70,  # F
        size_list[0],
        size_list[1],
        size_list[2],
        size_list[3],
        87,  # W
        69,  # E
        66,  # B
        80,  # P
        86,  # V
        80,  # P
        56,  # 8
    ]

    for bit in data:
        buffer.append(101 ^ bit)

    return bytes(buffer)


def do_login(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        server = args[0]
        if not server.logged_in:
            server.do_login()

        if server.logged_in:
            return func(*args, **kwargs)

        logger.warning('Server %s is not logged in. Please check credential in Settings', server.name)
        return None

    return wrapper


def get_allowed_servers_list(settings):
    servers_settings = settings.servers_settings
    servers_languages = settings.servers_languages

    servers = []
    for server_data in get_servers_list():
        if servers_languages and server_data['lang'] and server_data['lang'] not in servers_languages:
            continue

        server_settings = servers_settings.get(get_server_main_id_by_id(server_data['id']))
        if server_settings is not None:
            if not server_settings.get('enabled', True):
                # Server is disabled
                continue
            if len(servers_languages) > 0:
                if server_data['lang'] and server_data['lang'] not in servers_languages:
                    # Server language is not in list of enabled server languages
                    continue
                if not server_settings.get('langs', {}).get(server_data['lang'], True):
                    # Server is disabled for this language
                    continue

        if settings.nsfw_content is False and server_data['is_nsfw']:
            continue
        if settings.nsfw_only_content is False and server_data['is_nsfw_only']:
            continue

        servers.append(server_data)

    return servers


def get_server_class_name_by_id(id):
    """
    Returns a server class name from its ID

    `id` must respect the following format: `name[_lang][_whatever][:module_name]`

    + `name` is the name of the server.
    + `lang` is the language of the server (optional).
        Only useful when server belongs to a multi-languages server.
    + `whatever` is any string (optional).
        Only useful when a server must be backed up because it's dead.
        Beware, if `whatever` is defined, `lang` must be present even if it's empty.
        Example of value: old, bak, dead,...
    + `module_name` is the name of the module in which the server is defined (optional).
        Only useful if `module_name` is different from `name`.

    Parameters
    ----------
    :param id: A server ID
    :type id: str

    :return: The server class name corresponding to ID
    :rtype: str
    """
    return id.split(':')[0].capitalize()


def get_server_dir_name_by_id(id):
    name = id.split(':')[0]
    # Remove _whatever
    name = '_'.join(filter(None, name.split('_')[:2]))

    return name


def get_server_main_id_by_id(id):
    return id.split(':')[0].split('_')[0]


def get_server_module_name_by_id(id):
    return id.split(':')[-1].split('_')[0]


def get_servers_list(include_disabled=False, order_by=('lang', 'name')):
    count = 0
    servers = []
    for module in get_servers_modules():
        for _name, obj in dict(inspect.getmembers(module)).items():
            if not hasattr(obj, 'id') or not hasattr(obj, 'name') or not hasattr(obj, 'lang'):
                continue
            if NotImplemented in (obj.id, obj.name, obj.lang):
                continue

            if not include_disabled and obj.status == 'disabled':
                continue

            count += 1
            if inspect.isclass(obj):
                logo_path = os.path.join(get_cached_logos_dir(), 'servers', get_server_main_id_by_id(obj.id) + '.png')

                servers.append({
                    'id': obj.id,
                    'name': obj.name,
                    'content': obj.content,
                    'description': obj.description,
                    'lang': obj.lang,
                    'base_url': obj.base_url,
                    'has_login': obj.has_login,
                    'is_nsfw': obj.is_nsfw,
                    'is_nsfw_only': obj.is_nsfw_only,
                    'logo_path': logo_path if os.path.exists(logo_path) else None,
                    'logo_url': obj.logo_url,
                    'module': module,
                    'filters': obj.filters,
                    'params': obj.params,
                    'class_name': get_server_class_name_by_id(obj.id),
                })

    logger.info('Found %d servers', count)

    return sorted(servers, key=itemgetter(*order_by))


def get_servers_modules():
    def import_external_modules(servers_path, modules, modules_names, multi=False):
        if multi:
            servers_path = os.path.join(servers_path, 'multi')

        count = 0
        for path in glob.glob(os.path.join(servers_path, '*')):
            if not multi and os.path.isfile(path):
                continue

            name = os.path.basename(path)
            module_name = f'komikku.servers.multi.{name}' if multi else f'komikku.servers.{name}'
            if module_name in modules_names:
                continue

            if name == 'multi':
                continue

            module = importlib.import_module(f'.{name}', package='komikku.servers.multi' if multi else 'komikku.servers')
            modules.append(module)
            modules_names.append(module_name)
            count += 1

        if count > 0:
            logger.info('Import %d server modules from external folder %s', count, servers_path)

        return count

    def import_internal_modules(namespace, modules, modules_names):
        count = 0
        for _finder, module_name, ispkg in iter_namespace(namespace):
            if module_name in modules_names or not ispkg or module_name.endswith('.multi'):
                continue

            module = importlib.import_module(module_name)
            modules.append(module)
            modules_names.append(module_name)
            count += 1

        if count > 0:
            logger.info('Import %d server modules from internal folder', count)

        return count

    def iter_namespace(ns_pkg):
        # Specifying the second argument (prefix) to iter_modules makes the
        # returned name an absolute name instead of a relative one. This allows
        # import_module to work without having to do additional modification to
        # the name.
        return iter_modules(ns_pkg.__path__, ns_pkg.__name__ + '.')

    internal_done = False
    modules = []
    modules_names = []
    for finder in sys.meta_path:
        if isinstance(finder, ServerFinder):
            # Import servers from external folders
            for servers_path in finder.paths:
                if not os.path.exists(servers_path):
                    # Not very likely
                    continue

                import_external_modules(servers_path, modules, modules_names, multi=False)

        elif not internal_done:
            # Import internal servers
            import komikku.servers

            count = import_internal_modules(komikku.servers, modules, modules_names)

            if count > 0:
                internal_done = True

    return modules


def get_session_cookies(s):
    """
    Returns the cookies of a session
    regardless of the HTTP client (requests, curl_cffi)

    :param s: A session
    :type s: requests.sessions.Session or curl_cffi.requests.session.Session

    :return: a cookies Jar
    :rtype: requests.cookies.RequestsCookieJar or http.cookiejar.CookieJar
    """
    return s.cookies.jar if hasattr(s.cookies, 'jar') else s.cookies


def get_soup_element_inner_text(tag, text=None, sep=' ', recursive=True):
    """
    Returns inner text of a tag

    :param tag: A Tag
    :type tag: bs4.element.Tag

    :param text: A optional list of text strings to prepend
    :type text: list of str

    :param recursive: Recursively walk in children or not
    :type recursive: bool

    :return: The inner text of tag
    :rtype: str
    """
    if text is None:
        text = []

    for el in tag:
        if isinstance(el, NavigableString):
            text.append(el.strip())
        elif recursive:
            get_soup_element_inner_text(el, text)

    return sep.join(text).strip()


def parse_nextjs_hydration(soup, keyword, key=None):
    """
    Goes through all scripts in `soup`

    If a Next.js hydration containing `keyword` is found, it's deserialized and returned

    :param keyword: A string to identify the script being searched for
    :type text: str

    :param key: A key to retrieve a specific value rather than all data (optional)
    :type text: str
    """

    data = None

    for script_element in soup.select('script'):
        script = script_element.string
        if not script or not script.startswith('self.__next_f.push([1,') or f'\\"{keyword}\\":' not in script:  # noqa: E231, E701
            continue

        line = script.strip().replace('self.__next_f.push([1,', '')

        start = 0
        for c in line:
            if c in ('{', '['):
                break
            start += 1

        line = line[start:-3]

        try:
            data = json.loads(json.loads(f'"{line}"'))
        except Exception as e:
            logger.debug('ERROR: %s', line)
            logger.debug(e)
        break

    if key:
        key_value = None

        def get_key(data):
            nonlocal key_value

            if isinstance(data, list):
                for value in data:
                    get_key(value)

            elif isinstance(data, dict):
                for key, value in data.items():
                    if key != keyword:
                        get_key(value)
                        continue

                    key_value = value
                    break

        get_key(data)

        return key_value

    return data


def sojson4_decode(s):
    """
    Decodes a Sojson v4 string

    :param s: A string
    :type s: str

    :return: The decoded string
    :rtype: str
    """
    ss = re.split(r'[a-zA-Z]{1,}', s[240:-58])
    sss = ''
    for c in ss:
        sss += chr(int(c))

    return sss


def remove_emoji_from_string(text):
    """
    Removes Emojis from text (use emoji package)

    :param text: A text string
    :type text: str

    :return: The text string freed from Emojis
    :rtype: str
    """
    return emoji.replace_emoji(text, replace='').strip()


def search_duckduckgo(site, term, nb_pages=1):
    """
    Searches DuckDuckGo lite

    :param site: the site URL
    :type site: str

    :param term: the term to search for
    :type term: str

    :param nb_pages: the number of results pages to parse
    :type nb_pages: int

    :return: A list of dictionaries (name, url)
    :rtype: list of dict
    """
    base_url = 'https://lite.duckduckgo.com'

    session = requests.Session()
    session.headers.update({
        'Content-Type': 'application/x-www-form-urlencoded',
        'Origin': base_url,
        'Referer': f'{base_url}/',
        'User-Agent': USER_AGENT,
    })

    def extract_params(soup):
        """Extract vqd, s, dc params from HTML results page"""
        params = {}
        if input := soup.select_one('input[name="vqd"]'):
            params['vqd'] = input.get('value')
        if input := soup.select_one('input[name="s"]'):
            params['s'] = input.get('value')
        if input := soup.select_one('input[name="dc"]'):
            params['dc'] = input.get('value')

        return params

    def get_page(q, next_params):
        """Get and parse HTML results page"""
        params = {
            'q': q,
            'kl': '',
        }
        if next_params:
            params.update(next_params)
            params.update({
                'nextParams': '',
                'v': 'l',
                'o': 'json',
                'api': 'd.js',
                'kl': 'wt-wt',
            })
        else:
            params.update({
                'df': '',
            })

        r = session.post(base_url + '/lite/', data=params)
        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('.result-link'):
            url = a_element.get('href')
            if 'duckduckgo.com' in url:
                # Sponsored link
                continue

            results.append({
                'name': a_element.text.strip(),
                'url': a_element.get('href'),
            })

        return results, extract_params(soup)

    data = []
    next_params = None
    q = f'site:{site} {term}'  # noqa: E231
    for _index in range(nb_pages):
        results, next_params = get_page(q, next_params)
        data += results
        if not next_params:
            break
        time.sleep(1)

    return data


# https://github.com/Harkame/JapScanDownloader
def unscramble_image(image):
    """
    Unscrambles an image

    :param image: An image object
    :type image: PIL.Image.Image, bytes

    :return: A new unscrambled image
    :rtype: PIL.Image.Image
    """
    if not isinstance(image, Image.Image):
        image = Image.open(BytesIO(image))

    temp = Image.new('RGB', image.size)
    output_image = Image.new('RGB', image.size)

    for x in range(0, image.width, 200):
        col1 = image.crop((x, 0, x + 100, image.height))

        if x + 200 <= image.width:
            col2 = image.crop((x + 100, 0, x + 200, image.height))
            temp.paste(col1, (x + 100, 0))
            temp.paste(col2, (x, 0))
        else:
            col2 = image.crop((x + 100, 0, image.width, image.height))
            temp.paste(col1, (x, 0))
            temp.paste(col2, (x + 100, 0))

    for y in range(0, temp.height, 200):
        row1 = temp.crop((0, y, temp.width, y + 100))

        if y + 200 <= temp.height:
            row2 = temp.crop((0, y + 100, temp.width, y + 200))
            output_image.paste(row1, (0, y + 100))
            output_image.paste(row2, (0, y))
        else:
            row2 = temp.crop((0, y + 100, temp.width, temp.height))
            output_image.paste(row1, (0, y))
            output_image.paste(row2, (0, y + 100))

    return output_image


class RC4:
    """ RC4 (Rivest Cipher 4) stream cipher"""

    def __init__(self, key):
        self.key = key.encode()
        self.keystream = self.PRGA(self.KSA())

        # RC4-drop[256]
        for _i in range(256):
            next(self.keystream)

    def KSA(self):
        # Initialize S as a list of integers from 0 to 255
        S = list(range(256))
        j = 0

        for i in range(256):
            j = (j + S[i] + self.key[i % len(self.key)]) % 256
            # Swap values
            S[i], S[j] = S[j], S[i]

        return S

    def PRGA(self, S):
        i = j = 0

        while True:
            i = (i + 1) % 256
            j = (j + S[i]) % 256
            # Swap values
            S[i], S[j] = S[j], S[i]
            K = S[(S[i] + S[j]) % 256]
            yield K


class RC4SeedRandom:
    def __init__(self, key):
        self.key = key
        self.pos = 256

    def get_next(self):
        if self.pos == 256:
            self.keystream = RC4(self.key).keystream
            self.pos = 0
        self.pos += 1

        return next(self.keystream)


def unscramble_image_rc4(image, key, piece_size):
    """
    Unscrambles an image shuffled with RC4 stream cipher

    :param image: An image object
    :type image: PIL.Image.Image, bytes

    :param key: The key
    :type key: str

    :param piece_size: The pieces size
    :type piece_size: int

    :return: A new unscrambled image
    :rtype: PIL.Image.Image
    """
    if not isinstance(image, Image.Image):
        image = Image.open(BytesIO(image))

    output_image = Image.new('RGB', image.size)

    pieces = []
    for j in range(math.ceil(image.height / piece_size)):
        for i in range(math.ceil(image.width / piece_size)):
            pieces.append({
                'x': piece_size * i,
                'y': piece_size * j,
                'w': min(piece_size, image.width - piece_size * i),
                'h': min(piece_size, image.height - piece_size * j)
            })

    groups = {}
    for k, v in itertools.groupby(pieces, key=lambda x: x['w'] << 16 | x['h']):
        if k not in groups:
            groups[k] = []
        groups[k] += list(v)

    for _w, group in groups.items():
        size = len(group)

        permutation = []
        indexes = list(range(size))
        random = RC4SeedRandom(key)
        for i in range(size):
            num = random.get_next()
            exp = 8
            while num < (1 << 52):
                num = num << 8 | random.get_next()
                exp += 8
            while num >= (1 << 53):
                num = num >> 1
                exp -= 1

            permutation.append(indexes.pop(int(num * (2 ** -exp) * len(indexes))))

        for i, original in enumerate(permutation):
            src = group[i]
            dst = group[original]

            src_piece = image.crop((src['x'], src['y'], src['x'] + src['w'], src['y'] + src['h']))
            output_image.paste(src_piece, (dst['x'], dst['y']))

    return output_image



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.my_manga_reader_cms import MyMangaReaderCMS


class Scanfr(MyMangaReaderCMS):
    id = 'scanfr'
    name = 'Scan FR'
    lang = 'fr'
    status = 'disabled'  # 2024/03 gone

    base_url = 'https://www.scan-fr.org'
    search_url = base_url + '/search'
    most_populars_url = base_url + '/filterList?page=1&sortBy=views&asc=false'
    manga_url = base_url + '/manga/{0}'
    chapter_url = base_url + '/manga/{0}/{1}'
    image_url = None  # For some manga chapters, chapter slug can't be used to compute images URLs
    cover_url = 'https://opfrcdn.xyz/uploads/manga/{0}/cover/cover_250x350.jpg'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
import gc
from gettext import gettext as _
import threading

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import Gio
from gi.repository import Gtk
from gi.repository import Pango

from komikku.models import Settings
from komikku.servers import LANGUAGES
from komikku.utils import log_error_traceback

from komikku.explorer.common import ExplorerServerRow
from komikku.explorer.common import ExplorerSearchResultRow
from komikku.explorer.common import ExplorerSearchStackPage
from komikku.explorer.common import get_server_default_search_filters
from komikku.explorer.common import set_missing_server_logos


class ExplorerSearchStackPageSearchGlobal(ExplorerSearchStackPage):
    __gtype_name__ = 'ExplorerSearchStackPageSearchGlobal'

    lock = False
    selected_filters = []
    server_ids = []

    def __init__(self, parent):
        ExplorerSearchStackPage.__init__(self, parent)

        self.stack = self.parent.search_stack
        self.listbox = self.parent.search_listbox  # shared with explorer.search page
        self.filters_menu_button = self.parent.filters_menu_button

        self.selected_filters = Settings.get_default().explorer_search_global_selected_filters

        self.window.builder.add_from_resource('/info/febvre/Komikku/ui/menu/explorer_search_global_search.xml')
        self.filters_menu_button.set_menu_model(self.window.builder.get_object('menu-explorer-search-global-search'))

    def add_actions(self):
        action = Gio.SimpleAction.new_stateful(
            'explorer.search.global.search.pinned', None, GLib.Variant('b', 'pinned' in self.selected_filters)
        )
        action.connect('change-state', self.on_menu_action_changed)
        self.window.application.add_action(action)

    def cancel(self):
        def do_cancel():
            if self.status in ('ended', 'cancelled'):
                return False
            if self.status == 'started':
                return True

            self.status = 'cancelling'
            self.window.webview.cancel_challengers(self.server_ids, context='search')

        GLib.idle_add(do_cancel)

    def init_filters_menu(self):
        # Show filter menu button in searchbar
        self.filters_menu_button.set_visible(True)
        # Hide filters button in headerbar
        self.parent.filters_button.set_visible(False)

        if self.selected_filters:
            self.filters_menu_button.add_css_class('accent')
        else:
            self.filters_menu_button.remove_css_class('accent')

    def on_menu_action_changed(self, action, variant):
        value = variant.get_boolean()
        action.set_state(GLib.Variant('b', value))
        name = action.props.name.split('.')[-1]

        if value:
            self.selected_filters.add(name)
        else:
            self.selected_filters.remove(name)
        Settings.get_default().explorer_search_global_selected_filters = self.selected_filters

        if self.selected_filters:
            self.filters_menu_button.add_css_class('accent')
        else:
            self.filters_menu_button.remove_css_class('accent')

    def search(self, term):
        if self.lock:
            self.window.add_notification(_('A request is already in progress.'), timeout=2)
            return

        def run(servers):
            with ThreadPoolExecutor(max_workers=len(servers)) as executor:
                tasks = {}
                for server_data in servers:
                    future = executor.submit(search_server, server_data)
                    tasks[future] = server_data

                self.status = 'ongoing'

                for index, future in enumerate(as_completed(tasks)):
                    if self.window.page != self.parent.props.tag and self.window.previous_page != self.parent.props.tag:
                        executor.shutdown(False, cancel_futures=True)
                        break

                    server_data = tasks[future]
                    try:
                        results = future.result()
                    except Exception as exc:
                        GLib.idle_add(complete_server, None, server_data, log_error_traceback(exc))
                    else:
                        GLib.idle_add(complete_server, results, server_data)

                    self.parent.progressbar.set_fraction((index + 1) / len(servers))

            gc.collect()

            GLib.idle_add(complete)

        def complete():
            self.status = 'ended' if self.status == 'ongoing' else 'cancelled'
            self.lock = False
            self.parent.progressbar.set_fraction(0)

        def complete_server(results, server_data, message=None):
            server = getattr(server_data['module'], server_data['class_name'])()

            # Remove spinner
            for row in self.listbox:
                if row.server_data['lang'] == server.lang and row.server_data['name'] == server.name:
                    if row.position == 0:
                        row.results = results is not None and len(results) > 0
                    elif row.position == 1:
                        self.listbox.remove(row)
                        break

            if results:
                # Add results
                for index, item in enumerate(results):
                    row = ExplorerSearchResultRow(item)
                    row.server_data = server_data
                    row.position = index + 1
                    row.results = True
                    self.listbox.append(row)

                    if row.has_cover:
                        self.queue.put((row, server))
            else:
                # Error, cancelled or no results
                row = Gtk.ListBoxRow(activatable=False)
                row.server_data = server_data
                row.position = 1
                row.results = False
                row.is_result = False
                row.add_css_class('explorer-server-listboxrow')
                label = Gtk.Label(halign=Gtk.Align.CENTER, justify=Gtk.Justification.CENTER)
                if results is None:
                    if self.status == 'cancelling':
                        text = _('Search cancelled')
                    else:
                        # Error
                        text = _('Oops, search failed. Please try again.')
                        if message:
                            text = f'{text}\n{message}'
                else:
                    # No results
                    text = _('No results')
                label.set_markup(f'<i>{text}</i>')
                label.set_ellipsize(Pango.EllipsizeMode.END)
                row.set_child(label)

                self.listbox.append(row)

            self.listbox.invalidate_sort()

            self.render_covers()

        def search_server(server_data):
            server = getattr(server_data['module'], server_data['class_name'])()
            filters = get_server_default_search_filters(server)

            return server.search(term, **filters)

        def sort_results(row1, row2):
            """
            This function gets two children and has to return:
            - a negative integer if the first one should come before the second one
            - zero if they are equal
            - a positive integer if the second one should come before the firstone
            """
            row1_results = row1.results
            row1_server_lang = LANGUAGES.get(row1.server_data['lang'], '')
            row1_server_name = row1.server_data['name']
            row1_position = row1.position

            row2_results = row2.results
            row2_server_lang = LANGUAGES.get(row2.server_data['lang'], '')
            row2_server_name = row2.server_data['name']
            row2_position = row2.position

            # Servers with results first
            if row1_results and not row2_results:
                return -1
            if not row1_results and row2_results:
                return 1

            # Sort by language
            if row1_server_lang < row2_server_lang:
                return -1

            if row1_server_lang == row2_server_lang:
                # Sort by server name
                if row1_server_name < row2_server_name:
                    return -1

                # Sort by position
                if row1_server_name == row2_server_name and row1_position < row2_position:
                    return -1

            return 1

        self.clear()

        if 'pinned' in self.selected_filters:
            servers = []
            pinned_servers = Settings.get_default().pinned_servers
            for server_data in self.parent.parent.servers_page.servers:
                if server_data['id'] not in pinned_servers:
                    continue

                servers.append(server_data)
        else:
            servers = self.parent.parent.servers_page.servers

        # Init results list
        self.server_ids = []
        for server_data in servers:
            self.server_ids.append(server_data['id'])

            # Server row
            row = ExplorerServerRow(server_data, self.parent)
            row.server_data = server_data
            row.position = 0
            row.results = False
            row.is_result = False
            self.listbox.append(row)

            # Spinner
            row = Gtk.ListBoxRow(activatable=False)
            row.server_data = server_data
            row.position = 1
            row.results = False
            row.is_result = False
            row.add_css_class('explorer-server-listboxrow')
            spinner = Adw.Spinner()
            row.set_child(spinner)
            self.listbox.append(row)

        set_missing_server_logos(self.listbox)

        self.lock = True
        self.status = 'started'
        self.stack.set_visible_child_name('results')
        self.listbox.set_sort_func(sort_results)
        self.listbox.set_visible(True)

        thread = threading.Thread(target=run, args=(servers, ))
        thread.daemon = True
        thread.start()



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.pizzareader import PizzaReader


class Bluesolo(PizzaReader):
    id = 'bluesolo'
    name = 'Blue Solo'
    lang = 'fr'

    base_url = 'https://bluesolo.org'
    logo_url = base_url + '/storage/img/logo/Logo copie-72.png'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import datetime
from functools import cache
import json
import logging
import os
import shutil
import threading

from gi.repository import Gio
import natsort
import sqlite3

from komikku.utils import get_cached_data_dir
from komikku.utils import get_data_dir
from komikku.utils import is_flatpak

logger = logging.getLogger(__name__)

db_connections = threading.local()

VERSION = 15


def adapt_date_iso(val):
    """Adapt datetime.date to ISO 8601 date"""
    return val.isoformat()


def adapt_datetime_iso(val):
    """Adapt datetime.datetime to ISO 8601 datetime"""
    return val.isoformat()


def adapt_json(data):
    return (json.dumps(data, sort_keys=True)).encode()


def convert_date(val):
    """Convert ISO 8601 date to datetime.date object"""
    return datetime.date.fromisoformat(val.decode())


def convert_datetime(val):
    """Convert ISO 8601 datetime to datetime.datetime object"""
    return datetime.datetime.fromisoformat(val.decode())


def convert_json(blob):
    return json.loads(blob.decode())


sqlite3.register_adapter(datetime.date, adapt_date_iso)
sqlite3.register_adapter(datetime.datetime, adapt_datetime_iso)
sqlite3.register_adapter(dict, adapt_json)
sqlite3.register_adapter(list, adapt_json)
sqlite3.register_adapter(tuple, adapt_json)
sqlite3.register_converter('date', convert_date)
sqlite3.register_converter('json', convert_json)
sqlite3.register_converter('timestamp', convert_datetime)


def backup_db():
    db_path = get_db_path()

    if os.path.exists(db_path):
        db_conn = create_db_connection()
        if not db_conn:
            logger.info('[SQLite] Failed to backup DB')
            return

        if check_db(db_conn):
            logger.info('Save a DB backup')
            shutil.copyfile(db_path, get_db_backup_path())
        else:
            logger.info('[SQLite] Failed to backup DB')


def check_db(db_conn):
    try:
        res = db_conn.execute('PRAGMA integrity_check').fetchone()  # PRAGMA quick_check

        fk_violations = len(db_conn.execute('PRAGMA foreign_key_check').fetchall())

        ret = res[0] == 'ok' and fk_violations == 0
    except sqlite3.DatabaseError:
        logger.exception('[SQLite] Failed to check DB')
        ret = False

    return ret


def clear_cached_data(manga_in_use=None):
    # Clear chapters cache
    cache_dir_path = get_cached_data_dir()
    for server_name in os.listdir(cache_dir_path):
        server_dir_path = os.path.join(cache_dir_path, server_name)
        if manga_in_use and manga_in_use.path.startswith(server_dir_path):
            for manga_name in os.listdir(server_dir_path):
                manga_dir_path = os.path.join(server_dir_path, manga_name)
                if manga_dir_path != manga_in_use.path:
                    shutil.rmtree(manga_dir_path)
        else:
            shutil.rmtree(server_dir_path)

    # Clear database
    db_conn = create_db_connection()
    with db_conn:
        if manga_in_use:
            db_conn.execute('DELETE FROM mangas WHERE in_library != 1 AND id != ?', (manga_in_use.id, ))
        else:
            db_conn.execute('DELETE FROM mangas WHERE in_library != 1')


def collate_natsort(value1, value2):
    lst = natsort.natsorted([value1, value2], alg=natsort.ns.INT | natsort.ns.IC)
    return -1 if lst[0] == value1 else 1


def create_db_connection(path=None):
    if path is None:
        path = get_db_path()
    elif not os.path.exists(path):
        logger.error('[SQLite] Failed to create DB connection: invalid path %s', path)
        return None

    con = getattr(db_connections, path, None)

    needs_connect = False
    if con is None:
        needs_connect = True
        logger.debug(f'[SQLite] no DB connection exists for path {path} on thread {threading.get_native_id()}, creating new one')
    else:
        try:
            # Use fast query to check if the connection was closed
            con.execute('PRAGMA user_version')
        except sqlite3.ProgrammingError:
            logger.warn('[SQLite] DB connection was closed, reconnecting')
            needs_connect = True

    if needs_connect:
        con = sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES)
        if con is None:
            logger.error('[SQLite] Failed to create DB connection')
            return None

        con.row_factory = sqlite3.Row

        # Enable integrity constraint
        con.execute('PRAGMA foreign_keys = ON')

        # Ensure database can be accessed concurrently
        con.execute('PRAGMA journal_mode = WAL')

        # Perform less disk synchronization, safe under WAL
        con.execute('PRAGMA synchronous = NORMAL')
        con.execute('PRAGMA temp_store = MEMORY')

        # Add natural sort collation
        con.create_collation('natsort', collate_natsort)

        setattr(db_connections, path, con)

    return con


def execute_sql(conn, sql):
    res = True

    c = conn.cursor()

    try:
        c.execute(sql)
        conn.commit()
    except Exception as e:
        logger.error('[SQLite] %s', e)
        res = False

    c.close()

    return res


@cache
def get_db_path():
    app_profile = Gio.Application.get_default().profile

    if is_flatpak() and app_profile == 'beta':
        # In Flathub beta version share same data folder with stable version:
        # ~/.var/app/info.febvre.Komikku/data/
        # So, DB files must have distinct names
        name = 'komikku_beta.db'
    else:
        name = 'komikku.db'

    return os.path.join(get_data_dir(), name)


@cache
def get_db_backup_path():
    app_profile = Gio.Application.get_default().profile

    if is_flatpak() and app_profile == 'beta':
        name = 'komikku_beta_backup.db'
    else:
        name = 'komikku_backup.db'

    return os.path.join(get_data_dir(), name)


def init_db():
    db_conn = create_db_connection()
    if not db_conn:
        return

    db_path = get_db_path()
    db_backup_path = get_db_backup_path()

    if os.path.exists(db_path) and os.path.exists(db_backup_path) and not check_db(db_conn):
        # Restore backup
        logger.info('Restore DB from backup')
        shutil.copyfile(db_backup_path, db_path)

    sql_create_mangas_table = """CREATE TABLE IF NOT EXISTS mangas (
        id integer PRIMARY KEY,
        slug text NOT NULL,
        url text, -- only used in case slug can't be used to forge the url
        server_id text NOT NULL,
        in_library integer,
        name text NOT NULL,
        authors json,
        scanlators json,
        genres json,
        synopsis text,
        status text,
        background_color text,
        borders_crop integer,
        filters json,
        landscape_zoom integer,
        page_numbering integer,
        reading_mode text,
        scaling text,
        scaling_filter text,
        sort_order text,
        last_read timestamp,
        last_update timestamp,
        tracking json,
        UNIQUE (slug, server_id)
    );"""

    sql_create_chapters_table = """CREATE TABLE IF NOT EXISTS chapters (
        id integer PRIMARY KEY,
        manga_id integer REFERENCES mangas(id) ON DELETE CASCADE,
        slug text NOT NULL,
        url text, -- only used in case slug can't be used to forge the url
        title text NOT NULL,
        scanlators json,
        pages json,
        date date,
        num text,
        num_volume text,
        rank integer NOT NULL,
        downloaded integer NOT NULL,
        recent integer NOT NULL,
        read_progress text,
        read integer NOT NULL,
        last_page_read_index integer,
        last_read timestamp,
        UNIQUE (slug, manga_id)
    );"""

    sql_create_downloads_table = """CREATE TABLE IF NOT EXISTS downloads (
        id integer PRIMARY KEY,
        chapter_id integer REFERENCES chapters(id) ON DELETE CASCADE,
        status text NOT NULL,
        percent float NOT NULL,
        errors integer DEFAULT 0,
        date timestamp NOT NULL,
        UNIQUE (chapter_id)
    );"""

    sql_create_categories_table = """CREATE TABLE IF NOT EXISTS categories (
        id integer PRIMARY KEY,
        label text NOT NULL,
        UNIQUE (label)
    );"""

    sql_create_categories_mangas_association_table = """CREATE TABLE IF NOT EXISTS categories_mangas_association (
        category_id integer REFERENCES categories(id) ON DELETE CASCADE,
        manga_id integer REFERENCES mangas(id) ON DELETE CASCADE,
        UNIQUE (category_id, manga_id)
    );"""

    sql_create_indexes = [
        'CREATE INDEX IF NOT EXISTS idx_categories_mangas_association_manga ON categories_mangas_association(manga_id);',
        'CREATE INDEX IF NOT EXISTS idx_chapters_date ON chapters(manga_id, date);',
        'CREATE INDEX IF NOT EXISTS idx_chapters_downloaded ON chapters(manga_id, downloaded, read);',
        'CREATE INDEX IF NOT EXISTS idx_chapters_last_read ON chapters(last_read);',
        'CREATE INDEX IF NOT EXISTS idx_chapters_rank ON chapters(manga_id, rank);',
        'CREATE INDEX IF NOT EXISTS idx_chapters_read ON chapters(manga_id, read);',
        'CREATE INDEX IF NOT EXISTS idx_chapters_recent ON chapters(manga_id, recent);',
        'CREATE INDEX IF NOT EXISTS idx_downloads_date ON downloads(date);',
        'CREATE INDEX IF NOT EXISTS idx_downloads_status ON downloads(date, status);',
        'CREATE INDEX IF NOT EXISTS idx_mangas_in_library ON mangas(in_library, last_read DESC);',
    ]

    db_version = db_conn.execute('PRAGMA user_version').fetchone()[0]

    execute_sql(db_conn, sql_create_mangas_table)
    execute_sql(db_conn, sql_create_chapters_table)
    execute_sql(db_conn, sql_create_downloads_table)
    execute_sql(db_conn, sql_create_categories_table)
    execute_sql(db_conn, sql_create_categories_mangas_association_table)
    for sql_create_index in sql_create_indexes:
        execute_sql(db_conn, sql_create_index)

    if db_version == 0:
        # First launch
        db_conn.execute('PRAGMA user_version = {0}'.format(VERSION))

    if 0 < db_version <= 1:
        # Version 0.10.0
        if execute_sql(db_conn, 'ALTER TABLE downloads ADD COLUMN errors integer DEFAULT 0;'):
            db_conn.execute('PRAGMA user_version = {0}'.format(2))

    if 0 < db_version <= 2:
        # Version 0.12.0
        if execute_sql(db_conn, 'ALTER TABLE mangas ADD COLUMN borders_crop integer;'):
            db_conn.execute('PRAGMA user_version = {0}'.format(3))

    if 0 < db_version <= 4:
        # Version 0.16.0
        if execute_sql(db_conn, 'ALTER TABLE chapters ADD COLUMN scanlators json;'):
            db_conn.execute('PRAGMA user_version = {0}'.format(5))

    if 0 < db_version <= 5:
        # Version 0.22.0
        if execute_sql(db_conn, 'ALTER TABLE mangas RENAME COLUMN reading_direction TO reading_mode;'):
            db_conn.execute('PRAGMA user_version = {0}'.format(6))

    if 0 < db_version <= 6:
        # Version 0.25.0
        if execute_sql(db_conn, sql_create_categories_table) and execute_sql(db_conn, sql_create_categories_mangas_association_table):
            db_conn.execute('PRAGMA user_version = {0}'.format(7))

    if 0 < db_version <= 7:
        # Version 0.31.0
        ids_mapping = dict(
            jaiminisbox__old='jaiminisbox',
            kireicake='kireicake:jaiminisbox',
            lupiteam='lupiteam:jaiminisbox',
            tuttoanimemanga='tuttoanimemanga:jaiminisbox',

            readcomicsonline='readcomicsonline:hatigarmscans',

            hatigarmscans__old='hatigarmscans',

            edelgardescans='edelgardescans:genkan',
            hatigarmscans='hatigarmscans:genkan',
            hunlightscans='hunlightscans:genkan',
            leviatanscans__old='leviatanscans:genkan',
            leviatanscans_es_old='leviatanscans_es:genkan',
            oneshotscans__old='oneshotscans:genkan',
            reaperscans='reaperscans:genkan',
            thenonamesscans='thenonamesscans:genkan',
            zeroscans='zeroscans:genkan',

            akumanga='akumanga:madara',
            aloalivn='aloalivn:madara',
            apollcomics='apollcomics:madara',
            araznovel='araznovel:madara',
            argosscan='argosscan:madara',
            atikrost='atikrost:madara',
            romance24h='romance24h:madara',
            wakascan='wakascan:madara',
        )
        res = True
        for new, old in ids_mapping.items():
            res &= execute_sql(db_conn, f"UPDATE mangas SET server_id = '{new}' WHERE server_id = '{old}';")  # noqa: E702, E231

        if res:
            db_conn.execute('PRAGMA user_version = {0}'.format(8))

    if 0 < db_version <= 8:
        # Version 0.32.0
        if execute_sql(db_conn, 'ALTER TABLE mangas ADD COLUMN page_numbering integer;'):
            db_conn.execute('PRAGMA user_version = {0}'.format(9))

    if 0 < db_version <= 9:
        # Version 0.35.0
        execute_sql(db_conn, "UPDATE mangas SET server_id = 'reaperscans__old' WHERE server_id = 'reaperscans';")

        if execute_sql(db_conn, 'ALTER TABLE chapters ADD COLUMN last_read timestamp;'):
            db_conn.execute('PRAGMA user_version = {0}'.format(10))

    if 0 < db_version <= 10:
        # Version 1.0.0
        execute_sql(db_conn, 'ALTER TABLE mangas ADD COLUMN landscape_zoom integer;')
        execute_sql(db_conn, 'ALTER TABLE chapters ADD COLUMN read_progress text;')

        # Chapters: move reading status of pages in a new 'read_progress' field
        ids = []
        data = []
        manga_rows = db_conn.execute('SELECT id FROM mangas').fetchall()
        with db_conn:
            for manga_row in manga_rows:
                chapter_rows = db_conn.execute('SELECT * FROM chapters WHERE manga_id = ?', (manga_row['id'],)).fetchall()
                for chapter_row in chapter_rows:
                    if not chapter_row['pages']:
                        continue

                    read_progress = ''
                    for page in chapter_row['pages']:
                        read = page.pop('read', False)
                        read_progress += str(int(read))
                    if '1' in read_progress and '0' in read_progress:
                        ids.append(chapter_row['id'])
                        data.append({'pages': chapter_row['pages'], 'read_progress': read_progress})

            if ids:
                update_rows(db_conn, 'chapters', ids, data)

            db_conn.execute('PRAGMA user_version = {0}'.format(11))

    if 0 < db_version <= 11:
        # Version 1.16.0
        execute_sql(db_conn, 'ALTER TABLE mangas ADD COLUMN in_library integer;')
        execute_sql(db_conn, 'UPDATE mangas SET in_library = 1;')
        db_conn.execute('PRAGMA user_version = {0}'.format(12))

    if 0 < db_version <= 12:
        # Version 1.54.0
        execute_sql(db_conn, 'ALTER TABLE mangas ADD COLUMN scaling_filter text;')
        db_conn.execute('PRAGMA user_version = {0}'.format(13))

    if 0 < db_version <= 13:
        # Version 1.60.0
        execute_sql(db_conn, 'ALTER TABLE mangas ADD COLUMN tracking json;')
        execute_sql(db_conn, 'ALTER TABLE chapters ADD COLUMN num text;')
        execute_sql(db_conn, 'ALTER TABLE chapters ADD COLUMN num_volume text;')
        db_conn.execute('PRAGMA user_version = {0}'.format(14))

    if 0 < db_version <= 14:
        # Version 1.70.0
        execute_sql(db_conn, 'ALTER TABLE mangas ADD COLUMN filters json;')
        db_conn.execute('PRAGMA user_version = {0}'.format(15))

    logger.info('DB version {0}'.format(db_conn.execute('PRAGMA user_version').fetchone()[0]))


def delete_rows(db_conn, table, ids):
    seq = []
    if isinstance(ids[0], dict):
        # Several keys (secondary) are used to delete a row
        sql = 'DELETE FROM {0} WHERE {1}'.format(table, ' AND '.join(f'{skey} = ?' for skey in ids[0].keys()))

        for item in ids:
            seq.append(tuple(item.values()))
    else:
        sql = 'DELETE FROM {0} WHERE id = ?'.format(table)

        for id_ in ids:
            seq.append((id_, ))

    try:
        db_conn.executemany(sql, seq)
    except Exception as e:
        logger.error('[SQLite] %s %s', e, ids)
        return False
    else:
        return True


def insert_row(db_conn, table, data):
    try:
        cursor = db_conn.execute(
            'INSERT INTO {0} ({1}) VALUES ({2})'.format(table, ', '.join(data.keys()), ', '.join(['?'] * len(data))),
            tuple(data.values())
        )
    except Exception as e:
        logger.error('[SQLite] %s %s', e, data)
        return None
    else:
        return cursor.lastrowid


def insert_rows(db_conn, table, data):
    sql = 'INSERT INTO {0} ({1}) VALUES ({2})'.format(table, ', '.join(data[0].keys()), ', '.join(['?'] * len(data[0])))

    seq = []
    for item in data:
        seq.append(tuple(item.values()))

    try:
        db_conn.executemany(sql, seq)
    except Exception as e:
        logger.error('[SQLite] %s %s', e, data)
        return False
    else:
        return True


def update_row(db_conn, table, id_, data):
    try:
        db_conn.execute(
            'UPDATE {0} SET {1} WHERE id = ?'.format(table, ', '.join(k + ' = ?' for k in data)),
            tuple(data.values()) + (id_,)
        )
    except Exception as e:
        logger.error('[SQLite] %s %s', e, data)
        return False
    else:
        return True


def update_rows(db_conn, table, ids, data):
    sql = 'UPDATE {0} SET {1} WHERE id = ?'.format(table, ', '.join(k + ' = ?' for k in data[0]))

    seq = []
    for index, id_ in enumerate(ids):
        seq.append(tuple(data[index].values()) + (id_, ))

    try:
        db_conn.executemany(sql, seq)
    except Exception as e:
        logger.error('[SQLite] %s %s', e, data)
        return False
    else:
        return True



# SPDX-FileCopyrightText: 2025 Seth Falco
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Seth Falco <seth@falco.fun>

from gettext import gettext as _

from bs4 import BeautifulSoup

from komikku.servers.multi.hiveworks import Hiveworks
from komikku.utils import ServerContent


class Threepanelsoul(Hiveworks):
    id = 'threepanelsoul'
    name = 'Three Panel Soul'
    lang = 'en'
    content = ServerContent(
        type=[_('Webcomic'), _('Self-publishing')]
    )

    base_url = 'https://www.threepanelsoul.com'
    logo_url = base_url + '/favicon.ico'
    cover_url = None

    def get_metadata(self, soup: BeautifulSoup):
        return {
            'authors': [
                'Ian McConville',
                'Matthew Boyd',
            ],
            'synopsis': "It's a pretty rigid format but we keep the content loose, you know?",
        }
